import logging
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Any, Literal, cast

from databricks.labs.blueprint.installation import JsonValue
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.lakebridge.transpiler.transpile_status import TranspileError
from databricks.labs.lakebridge.reconcile.recon_config import Table


logger = logging.getLogger(__name__)


class LSPPromptMethod(Enum):
    FORCE = auto()  # for mandatory values that are specific to a dialect
    QUESTION = auto()
    CHOICE = auto()
    CONFIRM = auto()


@dataclass
class LSPConfigOptionV1:
    flag: str
    method: LSPPromptMethod
    prompt: str = ""
    choices: list[str] | None = None
    default: Any = None

    @classmethod
    def parse_all(cls, data: dict[str, Any]) -> dict[str, list["LSPConfigOptionV1"]]:
        return {key: list(LSPConfigOptionV1.parse(item) for item in value) for (key, value) in data.items()}

    @classmethod
    def parse(cls, data: Any) -> "LSPConfigOptionV1":
        if not isinstance(data, dict):
            raise ValueError(f"Invalid transpiler config option, expecting a dict entry, got {data}")
        flag: str = data.get("flag", "")
        if not flag:
            raise ValueError(f"Missing 'flag' entry in {data}")
        method_name: str = data.get("method", "")
        if not method_name:
            raise ValueError(f"Missing 'method' entry in {data}")
        method: LSPPromptMethod = cast(LSPPromptMethod, LSPPromptMethod[method_name])
        prompt: str = data.get("prompt", "")
        if not prompt:
            raise ValueError(f"Missing 'prompt' entry in {data}")
        choices = data.get("choices", [])
        default = data.get("default", None)
        return LSPConfigOptionV1(flag, method, prompt, choices, default)

    def prompt_for_value(self, prompts: Prompts) -> JsonValue:
        if self.method == LSPPromptMethod.FORCE:
            return self.default
        if self.method == LSPPromptMethod.CONFIRM:
            return prompts.confirm(self.prompt)
        if self.method == LSPPromptMethod.QUESTION:
            default = self.default if self.default else "None"
            result = prompts.question(self.prompt, default=default)
            if result == "<none>":
                return None
            return result
        if self.method == LSPPromptMethod.CHOICE:
            return prompts.choice(self.prompt, cast(list[str], self.choices))
        raise ValueError(f"Unsupported prompt method: {self.method}")


@dataclass
class TranspileConfig:
    __file__ = "config.yml"
    __version__ = 3

    transpiler_config_path: str | None = None
    source_dialect: str | None = None
    input_source: str | None = None
    output_folder: str | None = None
    error_file_path: str | None = None
    sdk_config: dict[str, str] | None = None
    skip_validation: bool = False
    catalog_name: str = "remorph"
    schema_name: str = "transpiler"
    transpiler_options: JsonValue = None

    @property
    def transpiler_path(self) -> Path | None:
        return Path(self.transpiler_config_path) if self.transpiler_config_path is not None else None

    @property
    def input_path(self) -> Path:
        if self.input_source is None:
            raise ValueError("Missing input source!")
        return Path(self.input_source)

    @property
    def output_path(self) -> Path | None:
        return None if self.output_folder is None else Path(self.output_folder)

    @property
    def error_path(self) -> Path | None:
        return Path(self.error_file_path) if self.error_file_path else None

    @property
    def target_dialect(self) -> Literal["databricks"]:
        return "databricks"

    @classmethod
    def v1_migrate(cls, raw: dict[str, Any]) -> dict[str, Any]:
        raw["version"] = 2
        return raw

    @classmethod
    def v2_migrate(cls, raw: dict[str, Any]) -> dict[str, Any]:
        del raw["mode"]
        key_mapping = {"input_sql": "input_source", "output_folder": "output_path", "source": "source_dialect"}
        raw["version"] = 3
        raw["error_file_path"] = "error_log.txt"
        return {key_mapping.get(key, key): value for key, value in raw.items()}


@dataclass
class TableRecon:
    __file__ = "recon_config.yml"
    __version__ = 1

    source_schema: str
    target_catalog: str
    target_schema: str
    tables: list[Table]
    source_catalog: str | None = None

    def __post_init__(self):
        self.source_schema = self.source_schema.lower()
        self.target_schema = self.target_schema.lower()
        self.target_catalog = self.target_catalog.lower()
        self.source_catalog = self.source_catalog.lower() if self.source_catalog else self.source_catalog


@dataclass
class DatabaseConfig:
    source_schema: str
    target_catalog: str
    target_schema: str
    source_catalog: str | None = None


@dataclass
class TranspileResult:
    transpiled_code: str
    success_count: int
    error_list: list[TranspileError]


@dataclass
class ValidationResult:
    validated_sql: str
    exception_msg: str | None


@dataclass
class ReconcileTablesConfig:
    filter_type: str  # all/include/exclude
    tables_list: list[str]  # [*, table1, table2]


@dataclass
class ReconcileMetadataConfig:
    catalog: str = "remorph"
    schema: str = "reconcile"
    volume: str = "reconcile_volume"


@dataclass
class ReconcileConfig:
    __file__ = "reconcile.yml"
    __version__ = 1

    data_source: str
    report_type: str
    secret_scope: str
    database_config: DatabaseConfig
    metadata_config: ReconcileMetadataConfig
    job_id: str | None = None
    tables: ReconcileTablesConfig | None = None


@dataclass
class LakebridgeConfiguration:
    transpile: TranspileConfig | None = None
    reconcile: ReconcileConfig | None = None



import tempfile
from pathlib import Path

from databricks.sdk.service.iam import User
from databricks.sdk.core import with_user_agent_extra

from databricks.labs.blueprint.entrypoint import get_logger
from databricks.labs.blueprint.tui import Prompts

from databricks.labs.bladespector.analyzer import Analyzer, _PLATFORM_TO_SOURCE_TECHNOLOGY

from databricks.labs.lakebridge.helpers.telemetry_utils import make_alphanum_or_semver
from databricks.labs.lakebridge.helpers.file_utils import check_path, move_tmp_file

logger = get_logger(__file__)


class LakebridgeAnalyzer(Analyzer):
    def __init__(self, current_user: User, prompts: Prompts, is_debug: bool = False):
        self._current_user = current_user
        self._prompts = prompts
        self._is_debug = is_debug
        super().__init__()

    def _get_source_directory(self) -> Path:
        """Get and validate the source directory from user input."""
        directory_str = self._prompts.question(
            "Enter full path to the source directory",
            default=Path.cwd().as_posix(),
            validate=check_path,
        )
        return Path(directory_str).resolve()

    def _get_result_file_path(self, directory: Path) -> Path:
        """Get the result file path - accepts either filename or full path."""
        filename = self._prompts.question(
            "Enter report file name or custom export path including file name without extension",
            default=f"{directory.as_posix()}/lakebridge-analyzer-results.xlsx",
            validate=check_path,
        )
        return directory / Path(filename) if len(filename.split("/")) == 1 else Path(filename)

    def _get_source_tech(self, platform: str | None = None) -> str:
        """Validate source technology or prompt for a valid source"""
        if platform is None or platform not in self.supported_source_technologies():
            if platform is not None:
                logger.warning(f"Invalid source technology {platform}")
            platform = self._prompts.choice("Select the source technology", self.supported_source_technologies())
            with_user_agent_extra("analyzer_source_tech", make_alphanum_or_semver(platform))
            logger.debug(f"User: {self._current_user}")
        return _PLATFORM_TO_SOURCE_TECHNOLOGY[platform]

    @staticmethod
    def _temp_xlsx_path(results_dir: Path | str) -> Path:
        return (Path(tempfile.mkdtemp()) / Path(results_dir).name).with_suffix(".xlsx")

    def _run_prompt_analyzer(self):
        """Run the analyzer: prompt guided"""
        source_dir = self._get_source_directory()
        results_dir = self._get_result_file_path(source_dir)
        tmp_dir = self._temp_xlsx_path(results_dir)
        technology = self._get_source_tech()

        self._run_binary(source_dir, tmp_dir, technology, self._is_debug)

        move_tmp_file(tmp_dir, results_dir)

        logger.info(f"Successfully Analyzed files in ${source_dir} for ${technology} and saved report to {results_dir}")

    def _run_arg_analyzer(self, source_dir: str | None, results_dir: str | None, technology: str | None):
        """Run the analyzer: arg guided"""
        if source_dir is None or results_dir is None or technology is None:
            logger.error("All arguments (--source-directory, --report-file, --source-tech) must be provided")
            return

        if check_path(source_dir) and check_path(results_dir):
            tmp_dir = self._temp_xlsx_path(results_dir)
            technology = self._get_source_tech(technology)
            self._run_binary(Path(source_dir), tmp_dir, technology, self._is_debug)

            move_tmp_file(tmp_dir, Path(results_dir))

            logger.info(
                f"Successfully Analyzed files in ${source_dir} for ${technology} and saved report to {results_dir}"
            )

    def run_analyzer(
        self, source_dir: str | None = None, results_dir: str | None = None, technology: str | None = None
    ):
        """Run the analyzer."""
        if not any([source_dir, results_dir, technology]):
            self._run_prompt_analyzer()
            return

        self._run_arg_analyzer(source_dir, results_dir, technology)



from abc import ABC, abstractmethod
import logging
import shutil
import yaml

from databricks.labs.blueprint.tui import Prompts

from databricks.labs.lakebridge.connections.credential_manager import (
    cred_file as creds,
    CredentialManager,
    create_credential_manager,
)
from databricks.labs.lakebridge.connections.database_manager import DatabaseManager
from databricks.labs.lakebridge.connections.env_getter import EnvGetter

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

PROFILER_SOURCE_SYSTEM = ["mssql", "synapse"]


class AssessmentConfigurator(ABC):
    """Abstract base class for assessment configuration."""

    def __init__(self, product_name: str, prompts: Prompts, credential_file=None):
        self.prompts = prompts
        self._product_name = product_name
        self._credential_file = creds(product_name) if not credential_file else credential_file

    @abstractmethod
    def _configure_credentials(self) -> str:
        pass

    @staticmethod
    def _test_connection(source: str, cred_manager: CredentialManager):
        config = cred_manager.get_credentials(source)

        try:
            db_manager = DatabaseManager(source, config)
            if db_manager.check_connection():
                logger.info("Connection to the source system successful")
            else:
                logger.error("Connection to the source system failed, check logs in debug mode")
                raise SystemExit("Connection validation failed. Exiting...")

        except ConnectionError as e:
            logger.error(f"Failed to connect to the source system: {e}")
            raise SystemExit("Connection validation failed. Exiting...") from e

    def run(self):
        """Run the assessment configuration process."""
        logger.info(f"Welcome to the {self._product_name} Assessment Configuration")
        source = self._configure_credentials()
        logger.info(f"{source.capitalize()} details and credentials received.")
        if self.prompts.confirm(f"Do you want to test the connection to {source}?"):
            cred_manager = create_credential_manager("lakebridge", EnvGetter())
            if cred_manager:
                self._test_connection(source, cred_manager)
        logger.info(f"{source.capitalize()} Assessment Configuration Completed")


class ConfigureSqlServerAssessment(AssessmentConfigurator):
    """SQL Server specific assessment configuration."""

    def _configure_credentials(self) -> str:
        cred_file = self._credential_file
        source = "mssql"

        logger.info(
            "\n(local | env) \nlocal means values are read as plain text \nenv means values are read "
            "from environment variables fall back to plain text if not variable is not found\n",
        )
        secret_vault_type = str(self.prompts.choice("Enter secret vault type (local | env)", ["local", "env"])).lower()
        secret_vault_name = None

        logger.info("Please refer to the documentation to understand the difference between local and env.")

        credential = {
            "secret_vault_type": secret_vault_type,
            "secret_vault_name": secret_vault_name,
            source: {
                "database": self.prompts.question("Enter the database name"),
                "driver": self.prompts.question("Enter the driver details"),
                "server": self.prompts.question("Enter the server or host details"),
                "port": int(self.prompts.question("Enter the port details", valid_number=True)),
                "user": self.prompts.question("Enter the user details"),
                "password": self.prompts.question("Enter the password details"),
            },
        }

        if cred_file.exists():
            backup_filename = cred_file.with_suffix('.bak')
            shutil.copy(cred_file, backup_filename)
            logger.debug(f"Backup of the existing file created at {backup_filename}")

        with open(cred_file, 'w', encoding='utf-8') as file:
            yaml.dump(credential, file, default_flow_style=False)

        logger.info(f"Credential template created for {source}.")
        return source


class ConfigureSynapseAssessment(AssessmentConfigurator):
    """Synapse specific assessment configuration."""

    def _configure_credentials(self) -> str:
        cred_file = self._credential_file
        source = "synapse"

        logger.info(
            "\n(local | env) \nlocal means values are read as plain text \nenv means values are read "
            "from environment variables fall back to plain text if not variable is not found\n",
        )
        secret_vault_type = str(self.prompts.choice("Enter secret vault type (local | env)", ["local", "env"])).lower()
        secret_vault_name = None

        # Synapse Workspace Settings
        logger.info("Please provide Synapse Workspace settings:")
        synapse_workspace = {
            "name": self.prompts.question("Enter Synapse workspace name"),
            "dedicated_sql_endpoint": self.prompts.question("Enter dedicated SQL endpoint"),
            "serverless_sql_endpoint": self.prompts.question("Enter serverless SQL endpoint"),
            "sql_user": self.prompts.question("Enter SQL user"),
            "sql_password": self.prompts.question("Enter SQL password"),
            "tz_info": self.prompts.question("Enter timezone (e.g. America/New_York)", default="UTC"),
        }

        # Azure API Access Settings
        logger.info("Please provide Azure API access settings:")
        azure_api_access = {
            "development_endpoint": self.prompts.question("Enter development endpoint"),
            "azure_client_id": self.prompts.question("Enter Azure client ID"),
            "azure_tenant_id": self.prompts.question("Enter Azure tenant ID"),
            "azure_client_secret": self.prompts.question("Enter Azure client secret"),
        }

        # JDBC Settings
        logger.info("Please select JDBC authentication type:")
        auth_type = self.prompts.choice(
            "Select authentication type", ["sql_authentication", "ad_passwd_authentication", "spn_authentication"]
        )

        synapse_jdbc = {
            "auth_type": auth_type,
            "fetch_size": self.prompts.question("Enter fetch size", default="1000"),
            "login_timeout": self.prompts.question("Enter login timeout (seconds)", default="30"),
        }

        # Profiler Settings
        logger.info("Please configure profiler settings:")
        synapse_profiler = {
            "exclude_serverless_sql_pool": self.prompts.confirm("Exclude serverless SQL pool from profiling?"),
            "exclude_dedicated_sql_pools": self.prompts.confirm("Exclude dedicated SQL pools from profiling?"),
            "exclude_spark_pools": self.prompts.confirm("Exclude Spark pools from profiling?"),
            "exclude_monitoring_metrics": self.prompts.confirm("Exclude monitoring metrics from profiling?"),
            "redact_sql_pools_sql_text": self.prompts.confirm("Redact SQL pools SQL text?"),
        }

        credential = {
            "secret_vault_type": secret_vault_type,
            "secret_vault_name": secret_vault_name,
            source: {
                "workspace": synapse_workspace,
                "azure_api_access": azure_api_access,
                "jdbc": synapse_jdbc,
                "profiler": synapse_profiler,
            },
        }

        if cred_file.exists():
            backup_filename = cred_file.with_suffix('.bak')
            shutil.copy(cred_file, backup_filename)
            logger.debug(f"Backup of the existing file created at {backup_filename}")

        with open(cred_file, 'w', encoding='utf-8') as file:
            yaml.dump(credential, file, default_flow_style=False)

        logger.info(f"Credential template created for {source}.")
        return source


def create_assessment_configurator(
    source_system: str, product_name: str, prompts: Prompts, credential_file=None
) -> AssessmentConfigurator:
    """Factory function to create the appropriate assessment configurator."""
    configurators = {
        "mssql": ConfigureSqlServerAssessment,
        "synapse": ConfigureSynapseAssessment,
    }

    if source_system not in configurators:
        raise ValueError(f"Unsupported source system: {source_system}")

    return configurators[source_system](product_name, prompts, credential_file)



from sqlglot import Dialects, Dialect

from databricks.labs.lakebridge.transpiler.sqlglot.parsers import oracle, presto, snowflake
from databricks.labs.lakebridge.transpiler.sqlglot.generator.databricks import Databricks

SQLGLOT_DIALECTS: dict[str, type[Dialect] | str] = {
    "athena": Dialects.ATHENA,
    "bigquery": Dialects.BIGQUERY,
    "databricks": Databricks,
    "mysql": Dialects.MYSQL,
    "netezza": Dialects.POSTGRES,
    "oracle": oracle.Oracle,
    "postgresql": Dialects.POSTGRES,
    "presto": presto.Presto,
    "redshift": Dialects.REDSHIFT,
    "snowflake": snowflake.Snowflake,
    "sqlite": Dialects.SQLITE,
    "teradata": Dialects.TERADATA,
    "trino": Dialects.TRINO,
    "tsql": Dialects.TSQL,
    "mssql": Dialects.TSQL,
    "synapse": Dialects.TSQL,
    "vertica": Dialects.POSTGRES,
}


def get_dialect(dialect: str) -> Dialect:
    return Dialect.get_or_raise(SQLGLOT_DIALECTS.get(dialect))


def get_key_from_dialect(input_dialect: Dialect) -> str:
    return [source_key for source_key, dialect in SQLGLOT_DIALECTS.items() if dialect == input_dialect][0]



import logging
import re
from importlib.resources import files

import databricks.labs.lakebridge.resources

from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.labs.lakebridge.helpers import db_sql

logger = logging.getLogger(__name__)


def replace_patterns(sql_text: str) -> str:
    """
    Replace the STRUCT and MAP datatypes in the SQL text with empty string
    """
    # Pattern to match nested STRUCT and MAP datatypes
    pattern = r'(STRUCT<[^<>]*?(?:<[^<>]*?>[^<>]*?)*>|MAP<[^<>]*?(?:<[^<>]*?>[^<>]*?)*>)'
    parsed_sql_text = re.sub(pattern, "", sql_text, flags=re.DOTALL)
    return parsed_sql_text


def extract_columns_with_datatype(sql_text: str) -> list[str]:
    """
    Extract the columns with datatype from the SQL text
    Example:
        Input: CREATE TABLE main (
            recon_table_id BIGINT NOT NULL,
            report_type STRING NOT NULL
            );
       Output:  [recon_table_id BIGINT NOT NULL,
                      report_type STRING NOT NULL]
    """
    return sql_text[sql_text.index("(") + 1 : sql_text.index(")")].strip().split(",")


def extract_column_name(column_with_datatype: str) -> str:
    """
    Extract the column name from the column with datatype.
    Example:
        Input: \n    recon_table_id BIGINT NOT NULL,
        Output: recon_table_id
    """
    return column_with_datatype.strip("\n").strip().split(" ")[0]


def table_original_query(table_name: str, full_table_name: str) -> str:
    """
    Get the main table DDL from the main.sql file
    :return: str
    """
    resources = files(databricks.labs.lakebridge.resources)
    query_dir = resources.joinpath("reconcile/queries/installation")
    return (
        query_dir.joinpath(f"{table_name}.sql")
        .read_text()
        .replace(f"CREATE TABLE IF NOT EXISTS {table_name}", f"CREATE OR REPLACE TABLE {full_table_name}")
    )


def current_table_columns(table_name: str, full_table_name: str) -> list[str]:
    """
    Extract the column names from the main table DDL
    :return: column_names: list[str]
    """
    main_sql = replace_patterns(table_original_query(table_name, full_table_name))
    main_table_columns = [
        extract_column_name(main_table_column) for main_table_column in extract_columns_with_datatype(main_sql)
    ]
    return main_table_columns


def installed_table_columns(ws: WorkspaceClient, table_identifier: str) -> list[str]:
    """
    Fetch the column names from the installed table on Databricks Workspace using SQL Backend
    :return: column_names: list[str]
    """
    main_table_columns = list(db_sql.get_sql_backend(ws).fetch(f"DESC {table_identifier}"))
    return [row.col_name for row in main_table_columns]


def check_table_mismatch(
    installed_table,
    current_table,
) -> bool:
    # Compare the current main table columns with the installed main table columns
    if len(installed_table) != len(current_table) or sorted(installed_table) != sorted(current_table):
        return True
    return False


def recreate_table_sql(
    table_identifier: str,
    installed_table: list[str],
    current_table: list[str],
    prompts: Prompts,
) -> str | None:
    """
    * Verify all the current main table columns are present in the installed main table and then use CTAS to recreate the main table
       * If any of the current main table columns are missing in the installed main table, prompt the user to recreate the main table:
            - If the user confirms, recreate the main table using the main DDL file, else log an error message and exit
    :param table_identifier:
    :param installed_table:
    :param current_table:
    :param prompts:
    :return:
    """
    table_name = table_identifier.split('.')[-1]
    sql: str | None = (
        f"CREATE OR REPLACE TABLE {table_identifier} AS SELECT {','.join(current_table)} FROM {table_identifier}"
    )

    if not set(current_table).issubset(installed_table):
        if prompts.confirm(
            f"The `{table_identifier}` table columns are not as expected. Do you want to recreate the `{table_identifier}` table?"
        ):
            sql = table_original_query(table_name, table_identifier)
        else:
            logger.error(
                f"The `{table_identifier}` table columns are not as expected. Please check and recreate the `{table_identifier}` table."
            )
            sql = None
    return sql



import logging
import re
from datetime import datetime

from pyspark.errors import PySparkException
from pyspark.sql import DataFrame, DataFrameReader, SparkSession
from pyspark.sql.functions import col
from sqlglot import Dialect
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.connectors.jdbc_reader import JDBCReaderMixin
from databricks.labs.lakebridge.reconcile.connectors.models import NormalizedIdentifier
from databricks.labs.lakebridge.reconcile.connectors.secrets import SecretsMixin
from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.exception import InvalidSnowflakePemPrivateKey
from databricks.labs.lakebridge.reconcile.recon_config import JdbcReaderOptions, Schema
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound

logger = logging.getLogger(__name__)


class SnowflakeDataSource(DataSource, SecretsMixin, JDBCReaderMixin):
    _DRIVER = "snowflake"
    _IDENTIFIER_DELIMITER = "\""

    """
       * INFORMATION_SCHEMA:
          - see https://docs.snowflake.com/en/sql-reference/info-schema#considerations-for-replacing-show-commands-with-information-schema-views
       * DATA:
          - only unquoted identifiers are treated as case-insensitive and are stored in uppercase.
          - for quoted identifiers refer:
             https://docs.snowflake.com/en/sql-reference/identifiers-syntax#double-quoted-identifiers
       * ORDINAL_POSITION:
          - indicates the sequential order of a column within a table or view,
             starting from 1 based on the order of column definition.
    """
    _SCHEMA_QUERY = """select column_name,
                                                      case
                                                            when numeric_precision is not null and numeric_scale is not null
                                                            then
                                                                concat(data_type, '(', numeric_precision, ',' , numeric_scale, ')')
                                                            when lower(data_type) = 'text'
                                                            then
                                                                concat('varchar', '(', CHARACTER_MAXIMUM_LENGTH, ')')
                                                            else data_type
                                                      end as data_type
                                                      from {catalog}.INFORMATION_SCHEMA.COLUMNS
                                                      where lower(table_name)='{table}' and table_schema = '{schema}'
                                                      order by ordinal_position"""

    def __init__(
        self,
        engine: Dialect,
        spark: SparkSession,
        ws: WorkspaceClient,
        secret_scope: str,
    ):
        self._engine = engine
        self._spark = spark
        self._ws = ws
        self._secret_scope = secret_scope

    @property
    def get_jdbc_url(self) -> str:
        try:
            sf_password = self._get_secret('sfPassword')
        except (NotFound, KeyError) as e:
            message = "sfPassword is mandatory for jdbc connectivity with Snowflake."
            logger.error(message)
            raise NotFound(message) from e

        return (
            f"jdbc:{SnowflakeDataSource._DRIVER}://{self._get_secret('sfAccount')}.snowflakecomputing.com"
            f"/?user={self._get_secret('sfUser')}&password={sf_password}"
            f"&db={self._get_secret('sfDatabase')}&schema={self._get_secret('sfSchema')}"
            f"&warehouse={self._get_secret('sfWarehouse')}&role={self._get_secret('sfRole')}"
        )

    def read_data(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        query: str,
        options: JdbcReaderOptions | None,
    ) -> DataFrame:
        table_query = query.replace(":tbl", f"{catalog}.{schema}.{table}")
        try:
            if options is None:
                df = self.reader(table_query).load()
            else:
                options = self._get_jdbc_reader_options(options)
                df = (
                    self._get_jdbc_reader(table_query, self.get_jdbc_url, SnowflakeDataSource._DRIVER)
                    .options(**options)
                    .load()
                )
            return df.select([col(column).alias(column.lower()) for column in df.columns])
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "data", table_query)

    def get_schema(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        normalize: bool = True,
    ) -> list[Schema]:
        """
        Fetch the Schema from the INFORMATION_SCHEMA.COLUMNS table in Snowflake.

        If the user's current role does not have the necessary privileges to access the specified
        Information Schema object, RunTimeError will be raised:
        "SQL access control error: Insufficient privileges to operate on schema 'INFORMATION_SCHEMA' "
        """
        schema_query = re.sub(
            r'\s+',
            ' ',
            SnowflakeDataSource._SCHEMA_QUERY.format(catalog=catalog, schema=schema.upper(), table=table),
        )
        try:
            logger.debug(f"Fetching schema using query: \n`{schema_query}`")
            logger.info(f"Fetching Schema: Started at: {datetime.now()}")
            df = self.reader(schema_query).load()
            schema_metadata = df.select([col(c).alias(c.lower()) for c in df.columns]).collect()
            logger.info(f"Schema fetched successfully. Completed at: {datetime.now()}")
            return [self._map_meta_column(field, normalize) for field in schema_metadata]
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "schema", schema_query)

    def reader(self, query: str) -> DataFrameReader:
        options = self._get_snowflake_options()
        return self._spark.read.format("snowflake").option("dbtable", f"({query}) as tmp").options(**options)

    # TODO cache this method using @functools.cache
    # Pay attention to https://pylint.pycqa.org/en/latest/user_guide/messages/warning/method-cache-max-size-none.html
    def _get_snowflake_options(self):
        options = {
            "sfUrl": self._get_secret('sfUrl'),
            "sfUser": self._get_secret('sfUser'),
            "sfDatabase": self._get_secret('sfDatabase'),
            "sfSchema": self._get_secret('sfSchema'),
            "sfWarehouse": self._get_secret('sfWarehouse'),
            "sfRole": self._get_secret('sfRole'),
        }
        options = options | self._get_snowflake_auth_options()

        return options

    def _get_snowflake_auth_options(self):
        try:
            key = SnowflakeDataSource._get_private_key(
                self._get_secret('pem_private_key'), self._get_secret_or_none('pem_private_key_password')
            )
            return {"pem_private_key": key}
        except (NotFound, KeyError):
            logger.warning("pem_private_key not found. Checking for sfPassword")
            try:
                password = self._get_secret('sfPassword')
                return {"sfPassword": password}
            except (NotFound, KeyError) as e:
                message = "sfPassword and pem_private_key not found. Either one is required for snowflake auth."
                logger.error(message)
                raise NotFound(message) from e

    @staticmethod
    def _get_private_key(pem_private_key: str, pem_private_key_password: str | None) -> str:
        try:
            private_key_bytes = pem_private_key.encode("UTF-8")
            password_bytes = pem_private_key_password.encode("UTF-8") if pem_private_key_password else None
        except UnicodeEncodeError as e:
            message = f"Invalid pem key and/or pem password: unable to encode. --> {e}"
            logger.error(message)
            raise ValueError(message) from e

        try:
            p_key = serialization.load_pem_private_key(
                private_key_bytes,
                password_bytes,
                backend=default_backend(),
            )
            pkb = p_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
            pkb_str = pkb.decode("UTF-8")
            # Remove the first and last lines (BEGIN/END markers)
            private_key_pem_lines = pkb_str.strip().split('\n')[1:-1]
            # Join the lines to form the base64 encoded string
            private_key_pem_str = ''.join(private_key_pem_lines)
            return private_key_pem_str
        except Exception as e:
            message = f"Failed to load or process the provided PEM private key. --> {e}"
            logger.error(message)
            raise InvalidSnowflakePemPrivateKey(message) from e

    def normalize_identifier(self, identifier: str) -> NormalizedIdentifier:
        return DialectUtils.normalize_identifier(
            identifier,
            source_start_delimiter=SnowflakeDataSource._IDENTIFIER_DELIMITER,
            source_end_delimiter=SnowflakeDataSource._IDENTIFIER_DELIMITER,
        )



import logging
from datetime import datetime
from uuid import uuid4

from pyspark.errors import PySparkException
from pyspark.sql import SparkSession

from databricks.sdk import WorkspaceClient

from databricks.labs.lakebridge.config import ReconcileConfig, TableRecon, DatabaseConfig
from databricks.labs.lakebridge.reconcile import utils
from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.exception import DataSourceRuntimeException, ReconciliationException
from databricks.labs.lakebridge.reconcile.recon_capture import (
    ReconCapture,
    ReconIntermediatePersist,
    generate_final_reconcile_output,
)
from databricks.labs.lakebridge.reconcile.recon_config import Table, Schema
from databricks.labs.lakebridge.reconcile.recon_output_config import (
    ReconcileOutput,
    ReconcileProcessDuration,
    SchemaReconcileOutput,
    DataReconcileOutput,
)
from databricks.labs.lakebridge.reconcile.reconciliation import Reconciliation
from databricks.labs.lakebridge.reconcile.schema_compare import SchemaCompare
from databricks.labs.lakebridge.reconcile.normalize_recon_config_service import NormalizeReconConfigService
from databricks.labs.lakebridge.transpiler.execute import verify_workspace_client
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_dialect

logger = logging.getLogger(__name__)
_RECON_REPORT_TYPES = {"schema", "data", "row", "all", "aggregate"}


class TriggerReconService:

    @staticmethod
    def trigger_recon(
        ws: WorkspaceClient,
        spark: SparkSession,
        table_recon: TableRecon,
        reconcile_config: ReconcileConfig,
        local_test_run: bool = False,
    ) -> ReconcileOutput:
        reconciler, recon_capture = TriggerReconService.create_recon_dependencies(
            ws, spark, reconcile_config, local_test_run
        )

        for table_conf in table_recon.tables:
            TriggerReconService.recon_one(spark, reconciler, recon_capture, reconcile_config, table_conf)

        return TriggerReconService.verify_successful_reconciliation(
            generate_final_reconcile_output(
                recon_id=recon_capture.recon_id,
                spark=spark,
                metadata_config=reconcile_config.metadata_config,
                local_test_run=local_test_run,
            )
        )

    @staticmethod
    def create_recon_dependencies(
        ws: WorkspaceClient, spark: SparkSession, reconcile_config: ReconcileConfig, local_test_run: bool = False
    ) -> tuple[Reconciliation, ReconCapture]:
        ws_client: WorkspaceClient = verify_workspace_client(ws)

        # validate the report type
        report_type = reconcile_config.report_type.lower()
        logger.info(f"report_type: {report_type}, data_source: {reconcile_config.data_source} ")
        utils.validate_input(report_type, _RECON_REPORT_TYPES, "Invalid report type")

        source, target = utils.initialise_data_source(
            engine=reconcile_config.data_source,
            spark=spark,
            ws=ws_client,
            secret_scope=reconcile_config.secret_scope,
        )

        recon_id = str(uuid4())
        # initialise the Reconciliation
        reconciler = Reconciliation(
            source,
            target,
            reconcile_config.database_config,
            report_type,
            SchemaCompare(spark=spark),
            get_dialect(reconcile_config.data_source),
            spark,
            metadata_config=reconcile_config.metadata_config,
        )

        recon_capture = ReconCapture(
            database_config=reconcile_config.database_config,
            recon_id=recon_id,
            report_type=report_type,
            source_dialect=get_dialect(reconcile_config.data_source),
            ws=ws_client,
            spark=spark,
            metadata_config=reconcile_config.metadata_config,
            local_test_run=local_test_run,
        )

        return reconciler, recon_capture

    @staticmethod
    def recon_one(
        spark: SparkSession,
        reconciler: Reconciliation,
        recon_capture: ReconCapture,
        reconcile_config: ReconcileConfig,
        table_conf: Table,
    ):
        normalized_table_conf = NormalizeReconConfigService(
            reconciler.source, reconciler.target
        ).normalize_recon_table_config(table_conf)

        schema_reconcile_output, data_reconcile_output, recon_process_duration = TriggerReconService._do_recon_one(
            reconciler, reconcile_config, normalized_table_conf
        )

        TriggerReconService.persist_delta_table(
            spark,
            reconciler,
            recon_capture,
            schema_reconcile_output,
            data_reconcile_output,
            reconcile_config,
            normalized_table_conf,
            recon_process_duration,
        )

    @staticmethod
    def _do_recon_one(reconciler: Reconciliation, reconcile_config: ReconcileConfig, table_conf: Table):
        recon_process_duration = ReconcileProcessDuration(start_ts=str(datetime.now()), end_ts=None)
        schema_reconcile_output = SchemaReconcileOutput(is_valid=True)
        data_reconcile_output = DataReconcileOutput()

        try:
            src_schema, tgt_schema = TriggerReconService.get_schemas(
                reconciler.source, reconciler.target, table_conf, reconcile_config.database_config, True
            )
        except DataSourceRuntimeException as e:
            schema_reconcile_output = SchemaReconcileOutput(is_valid=False, exception=str(e))
        else:
            if reconciler.report_type in {"schema", "all"}:
                schema_reconcile_output = TriggerReconService._run_reconcile_schema(
                    reconciler=reconciler,
                    table_conf=table_conf,
                    src_schema=src_schema,
                    tgt_schema=tgt_schema,
                )
                logger.warning("Schema comparison is completed.")

            if reconciler.report_type in {"data", "row", "all"}:
                data_reconcile_output = TriggerReconService._run_reconcile_data(
                    reconciler=reconciler,
                    table_conf=table_conf,
                    src_schema=src_schema,
                    tgt_schema=tgt_schema,
                )
                logger.warning(f"Reconciliation for '{reconciler.report_type}' report completed.")

        recon_process_duration.end_ts = str(datetime.now())
        return schema_reconcile_output, data_reconcile_output, recon_process_duration

    @staticmethod
    def get_schemas(
        source: DataSource,
        target: DataSource,
        table_conf: Table,
        database_config: DatabaseConfig,
        normalize: bool,
    ) -> tuple[list[Schema], list[Schema]]:
        src_schema = source.get_schema(
            catalog=database_config.source_catalog,
            schema=database_config.source_schema,
            table=table_conf.source_name,
            normalize=normalize,
        )

        tgt_schema = target.get_schema(
            catalog=database_config.target_catalog,
            schema=database_config.target_schema,
            table=table_conf.target_name,
            normalize=normalize,
        )

        return src_schema, tgt_schema

    @staticmethod
    def _run_reconcile_schema(
        reconciler: Reconciliation,
        table_conf: Table,
        src_schema: list[Schema],
        tgt_schema: list[Schema],
    ):
        try:
            return reconciler.reconcile_schema(table_conf=table_conf, src_schema=src_schema, tgt_schema=tgt_schema)
        except PySparkException as e:
            return SchemaReconcileOutput(is_valid=False, exception=str(e))

    @staticmethod
    def _run_reconcile_data(
        reconciler: Reconciliation,
        table_conf: Table,
        src_schema: list[Schema],
        tgt_schema: list[Schema],
    ) -> DataReconcileOutput:
        try:
            return reconciler.reconcile_data(table_conf=table_conf, src_schema=src_schema, tgt_schema=tgt_schema)
        except DataSourceRuntimeException as e:
            return DataReconcileOutput(exception=str(e))

    @staticmethod
    def persist_delta_table(
        spark: SparkSession,
        reconciler: Reconciliation,
        recon_capture: ReconCapture,
        schema_reconcile_output: SchemaReconcileOutput,
        data_reconcile_output: DataReconcileOutput,
        reconcile_config: ReconcileConfig,
        table_conf: Table,
        recon_process_duration: ReconcileProcessDuration,
    ):
        recon_capture.start(
            data_reconcile_output=data_reconcile_output,
            schema_reconcile_output=schema_reconcile_output,
            table_conf=table_conf,
            recon_process_duration=recon_process_duration,
            record_count=reconciler.get_record_count(table_conf, reconciler.report_type),
        )
        if reconciler.report_type != "schema":
            ReconIntermediatePersist(
                spark=spark, path=utils.generate_volume_path(table_conf, reconcile_config.metadata_config)
            ).clean_unmatched_df_from_volume()

    @staticmethod
    def verify_successful_reconciliation(
        reconcile_output: ReconcileOutput, operation_name: str = "reconcile"
    ) -> ReconcileOutput:
        for table_output in reconcile_output.results:
            if table_output.exception_message or (
                table_output.status.column is False
                or table_output.status.row is False
                or table_output.status.schema is False
                or table_output.status.aggregate is False
            ):
                raise ReconciliationException(
                    f" Reconciliation failed for one or more tables. Please check the recon metrics for more details."
                    f" **{operation_name}** failed.",
                    reconcile_output=reconcile_output,
                )

        logger.info("Reconciliation completed successfully.")
        return reconcile_output



import logging
from datetime import timedelta
from pathlib import Path

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.lsql.dashboards import DashboardMetadata, Dashboards
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import (
    InvalidParameterValue,
    NotFound,
    DeadlineExceeded,
    InternalError,
    ResourceAlreadyExists,
)
from databricks.sdk.retries import retried
from databricks.sdk.service.dashboards import LifecycleState, Dashboard

from databricks.labs.lakebridge.config import ReconcileConfig, ReconcileMetadataConfig

logger = logging.getLogger(__name__)


class DashboardDeployment:

    def __init__(
        self,
        ws: WorkspaceClient,
        installation: Installation,
        install_state: InstallState,
    ):
        self._ws = ws
        self._installation = installation
        self._install_state = install_state

    def deploy(
        self,
        folder: Path,
        config: ReconcileConfig,
    ):
        """
        Create dashboards from Dashboard metadata files.
        The given folder is expected to contain subfolders each containing metadata for individual dashboards.

        :param folder: Path to the base folder.
        :param config: Configuration for reconciliation.
        """
        logger.info(f"Deploying dashboards from base folder {folder}")
        parent_path = f"{self._installation.install_folder()}/dashboards"
        try:
            self._ws.workspace.mkdirs(parent_path)
        except ResourceAlreadyExists:
            logger.info(f"Dashboard parent path already exists: {parent_path}")

        valid_dashboard_refs = set()
        for dashboard_folder in folder.iterdir():
            # Make sure the directory contains a dashboard
            if not (dashboard_folder.is_dir() and dashboard_folder.joinpath("dashboard.yml").exists()):
                continue
            valid_dashboard_refs.add(self._dashboard_reference(dashboard_folder))
            dashboard = self._update_or_create_dashboard(dashboard_folder, parent_path, config.metadata_config)
            logger.info(
                f"Dashboard deployed with URL: {self._ws.config.host}/sql/dashboardsv3/{dashboard.dashboard_id}"
            )
            self._install_state.save()

        self._remove_deprecated_dashboards(valid_dashboard_refs)

    def _dashboard_reference(self, folder: Path) -> str:
        return f"{folder.stem}".lower()

    # InternalError and DeadlineExceeded are retried because of Lakeview internal issues
    # These issues have been reported to and are resolved by the Lakeview team
    # Keeping the retry for resilience
    @retried(on=[InternalError, DeadlineExceeded], timeout=timedelta(minutes=3))
    def _update_or_create_dashboard(
        self,
        folder: Path,
        ws_parent_path: str,
        config: ReconcileMetadataConfig,
    ) -> Dashboard:
        logging.info(f"Reading dashboard folder {folder}")
        metadata = DashboardMetadata.from_path(folder).replace_database(
            catalog=config.catalog,
            catalog_to_replace="remorph",
            database=config.schema,
            database_to_replace="reconcile",
        )

        metadata.display_name = self._name_with_prefix(metadata.display_name)
        reference = self._dashboard_reference(folder)
        dashboard_id = self._install_state.dashboards.get(reference)
        if dashboard_id is not None:
            try:
                dashboard_id = self._handle_existing_dashboard(dashboard_id, metadata.display_name)
            except (NotFound, InvalidParameterValue):
                logger.info(f"Recovering invalid dashboard: {metadata.display_name} ({dashboard_id})")
                try:
                    dashboard_path = f"{ws_parent_path}/{metadata.display_name}.lvdash.json"
                    self._ws.workspace.delete(dashboard_path)  # Cannot recreate dashboard if file still exists
                    logger.debug(
                        f"Deleted dangling dashboard {metadata.display_name} ({dashboard_id}): {dashboard_path}"
                    )
                except NotFound:
                    pass
                dashboard_id = None  # Recreate the dashboard if it's reference is corrupted (manually)

        dashboard = Dashboards(self._ws).create_dashboard(
            metadata,
            dashboard_id=dashboard_id,
            parent_path=ws_parent_path,
            warehouse_id=self._ws.config.warehouse_id,
            publish=True,
        )
        assert dashboard.dashboard_id is not None
        self._install_state.dashboards[reference] = dashboard.dashboard_id
        return dashboard

    def _name_with_prefix(self, name: str) -> str:
        prefix = self._installation.product()
        return f"{prefix.upper()}_{name}".replace(" ", "_")

    def _handle_existing_dashboard(self, dashboard_id: str, display_name: str) -> str | None:
        dashboard = self._ws.lakeview.get(dashboard_id)
        if dashboard.lifecycle_state is None:
            raise NotFound(f"Dashboard life cycle state: {display_name} ({dashboard_id})")
        if dashboard.lifecycle_state == LifecycleState.TRASHED:
            logger.info(f"Recreating trashed dashboard: {display_name} ({dashboard_id})")
            return None  # Recreate the dashboard if it is trashed (manually)
        return dashboard_id  # Update the existing dashboard

    def _remove_deprecated_dashboards(self, valid_dashboard_refs: set[str]):
        for ref, dashboard_id in self._install_state.dashboards.items():
            if ref not in valid_dashboard_refs:
                try:
                    logger.info(f"Removing dashboard_id={dashboard_id}, as it is no longer needed.")
                    del self._install_state.dashboards[ref]
                    self._ws.lakeview.trash(dashboard_id)
                except (InvalidParameterValue, NotFound):
                    logger.warning(f"Dashboard `{dashboard_id}` doesn't exist anymore for some reason.")
                    continue



from __future__ import annotations

import abc
import asyncio
import logging
import os
import shutil
import sys
import venv
from collections.abc import Callable, Sequence, Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal

import attrs
import yaml
from lsprotocol import types as types_module
from lsprotocol.types import (
    CLIENT_REGISTER_CAPABILITY,
    METHOD_TO_TYPES,
    ClientCapabilities,
    ClientInfo,
    Diagnostic,
    DiagnosticSeverity,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
    InitializeParams,
    InitializeResult,
    LanguageKind,
)
from lsprotocol.types import Position as LSPPosition
from lsprotocol.types import Range as LSPRange
from lsprotocol.types import Registration, RegistrationParams, TextDocumentIdentifier, TextDocumentItem, TextEdit
from pygls.exceptions import FeatureRequestError
from pygls.lsp.client import BaseLanguageClient

from databricks.labs.blueprint.wheels import ProductInfo
from databricks.labs.lakebridge.config import LSPConfigOptionV1, TranspileConfig, TranspileResult
from databricks.labs.lakebridge.errors.exceptions import IllegalStateException
from databricks.labs.lakebridge.helpers.file_utils import is_dbt_project_file, is_sql_file
from databricks.labs.lakebridge.transpiler.transpile_engine import TranspileEngine
from databricks.labs.lakebridge.transpiler.transpile_status import (
    CodePosition,
    CodeRange,
    ErrorKind,
    ErrorSeverity,
    TranspileError,
)

logger = logging.getLogger(__name__)


@dataclass
class _LSPRemorphConfigV1:
    name: str
    dialects: Sequence[str]
    env_vars: Mapping[str, str]
    command_line: Sequence[str]

    @classmethod
    def parse(cls, data: Mapping[str, Any]) -> _LSPRemorphConfigV1:
        version = data.get("version", 0)
        if version != 1:
            raise ValueError(f"Unsupported transpiler config version: {version}")
        name: str | None = data.get("name", None)
        if not name:
            raise ValueError("Missing 'name' entry")
        dialects = data.get("dialects", [])
        if len(dialects) == 0:
            raise ValueError("Missing 'dialects' entry")
        env_vars = data.get("environment", {})
        command_line = data.get("command_line", [])
        if len(command_line) == 0:
            raise ValueError("Missing 'command_line' entry")
        return _LSPRemorphConfigV1(name, dialects, env_vars, command_line)


@dataclass
class LSPConfig:
    path: Path
    remorph: _LSPRemorphConfigV1
    options: Mapping[str, Sequence[LSPConfigOptionV1]]
    custom: Mapping[str, Any]

    @property
    def name(self):
        return self.remorph.name

    def options_for_dialect(self, source_dialect: str) -> Sequence[LSPConfigOptionV1]:
        return [*self.options.get("all", []), *self.options.get(source_dialect, [])]

    @classmethod
    def load(cls, path: Path) -> LSPConfig:
        yaml_text = path.read_text()
        data = yaml.safe_load(yaml_text)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid transpiler config, expecting a dict, got a {type(data).__name__}")
        remorph_data = data.get("remorph", None)
        if not isinstance(remorph_data, dict):
            raise ValueError(f"Invalid transpiler config, expecting a 'remorph' dict entry, got {remorph_data}")
        remorph = _LSPRemorphConfigV1.parse(remorph_data)
        options_data = data.get("options", {})
        if not isinstance(options_data, dict):
            raise ValueError(f"Invalid transpiler config, expecting an 'options' dict entry, got {options_data}")
        options = LSPConfigOptionV1.parse_all(options_data)
        custom = data.get("custom", {})
        return LSPConfig(path, remorph, options, custom)


def lsp_feature(name: str, options: Any | None = None):
    def wrapped(func: Callable):
        _LSP_FEATURES.append((name, options, func))
        return func

    return wrapped


_LSP_FEATURES: list[tuple[str, Any | None, Callable]] = []

# the below code also exists in lsp_server.py
# it will be factorized as part of https://github.com/databrickslabs/remorph/issues/1304
TRANSPILE_TO_DATABRICKS_METHOD = "document/transpileToDatabricks"


@attrs.define
class TranspileDocumentParams:
    uri: str = attrs.field()
    language_id: LanguageKind | str = attrs.field()


@attrs.define
class TranspileDocumentRequest:
    # 'id' is mandated by LSP
    # pylint: disable=invalid-name
    id: int | str = attrs.field()
    params: TranspileDocumentParams = attrs.field()
    method: Literal["document/transpileToDatabricks"] = "document/transpileToDatabricks"
    jsonrpc: str = attrs.field(default="2.0")


@attrs.define
class TranspileDocumentResult:
    uri: str = attrs.field()
    language_id: LanguageKind | str = attrs.field()
    changes: Sequence[TextEdit] = attrs.field()
    diagnostics: Sequence[Diagnostic] = attrs.field()


@attrs.define
class TranspileDocumentResponse:
    # 'id' is mandated by LSP
    # pylint: disable=invalid-name
    id: int | str = attrs.field()
    result: TranspileDocumentResult = attrs.field()
    jsonrpc: str = attrs.field(default="2.0")


def install_special_properties():
    is_special_property = getattr(types_module, "is_special_property")

    def customized(cls: type, property_name: str) -> bool:
        if cls is TranspileDocumentRequest and property_name in {"method", "jsonrpc"}:
            return True
        return is_special_property(cls, property_name)

    setattr(types_module, "is_special_property", customized)


install_special_properties()

METHOD_TO_TYPES[TRANSPILE_TO_DATABRICKS_METHOD] = (
    TranspileDocumentRequest,
    TranspileDocumentResponse,
    TranspileDocumentParams,
    None,
)


# subclass BaseLanguageClient so we can override stuff when required
class _LanguageClient(BaseLanguageClient):

    def __init__(self, name: str, version: str) -> None:
        super().__init__(name, version)
        self._transpile_to_databricks_capability: Registration | None = None
        self._register_lsp_features()

    @property
    def is_alive(self):
        return self._server and self._server.returncode is None

    @property
    def transpile_to_databricks_capability(self):
        return self._transpile_to_databricks_capability

    @lsp_feature(CLIENT_REGISTER_CAPABILITY)
    def register_capabilities(self, params: RegistrationParams) -> None:
        for registration in params.registrations:
            if registration.method == TRANSPILE_TO_DATABRICKS_METHOD:
                logger.debug(f"Registered capability: {registration.method}")
                self._transpile_to_databricks_capability = registration
                continue
            logger.debug(f"Unknown capability: {registration.method}")

    async def transpile_document_async(self, params: TranspileDocumentParams) -> TranspileDocumentResult:
        """Transpile a document to Databricks SQL.

        The caller is responsible for ensuring that the LSP server is capable of handling this request.

        Args:
            params: The parameters for the transpile request to forward to the LSP server.
        Returns:
            The result of the transpile request, from the LSP server.
        Raises:
            IllegalStateException: If the client has been stopped or the server hasn't (yet) signalled that it is
                capable of transpiling documents to Databricks SQL.
        """
        if self.stopped:
            raise IllegalStateException("Client has been stopped.")
        if not self.transpile_to_databricks_capability:
            raise IllegalStateException("Client has not yet registered its transpile capability.")
        return await self.protocol.send_request_async(TRANSPILE_TO_DATABRICKS_METHOD, params)

    # can't use @client.feature because it requires a global instance
    def _register_lsp_features(self):
        for name, options, func in _LSP_FEATURES:
            decorator = self.protocol.fm.feature(name, options)
            wrapper = self._wrapper_for_lsp_feature(func)
            decorator(wrapper)

    def _wrapper_for_lsp_feature(self, func):
        def wrapper(params):
            return func(self, params)

        return wrapper

    async def start_io(self, cmd: str, *args, **kwargs):
        await super().start_io(cmd, *args, **kwargs)
        # forward stderr
        task = asyncio.create_task(self.pipe_stderr())
        self._async_tasks.append(task)

    async def pipe_stderr(self) -> None:
        server = self._server
        assert server is not None
        stderr = server.stderr
        assert stderr is not None
        while not self._stop_event.is_set():
            data: bytes = await stderr.readline()
            if not data:
                return
            # Invalid UTF-8 isn't great, but we can at least log it with the replacement character rather
            # than dropping it silently or triggering an exception.
            message = data.decode("utf-8", errors="replace").strip()
            # Although information may arrive via stderr, it's generally informational in nature and doesn't
            # necessarily represent an error
            # TODO: analyze message and log it accordingly (info/war/error...).
            logger.info(message)
            if not data.endswith(b"\n"):
                break


class ChangeManager(abc.ABC):

    @classmethod
    def apply(
        cls, source_code: str, changes: Sequence[TextEdit], diagnostics: Sequence[Diagnostic], file_path: Path
    ) -> TranspileResult:
        if not changes and not diagnostics:
            return TranspileResult(source_code, 1, [])
        transpile_errors = [DiagnosticConverter.apply(file_path, diagnostic) for diagnostic in diagnostics]
        try:
            lines = source_code.split("\n")
            for change in changes:
                lines = cls._apply(lines, change)
            transpiled_code = "\n".join(lines)
            return TranspileResult(transpiled_code, 1, transpile_errors)
        except IndexError as e:
            logger.error("Failed to apply changes", exc_info=e)
            error = TranspileError(
                code="INTERNAL_ERROR",
                kind=ErrorKind.INTERNAL,
                severity=ErrorSeverity.ERROR,
                path=file_path,
                message="Internal error, failed to apply changes",
            )
            transpile_errors.append(error)
            return TranspileResult(source_code, 1, transpile_errors)

    @classmethod
    def _apply(cls, lines: list[str], change: TextEdit) -> list[str]:
        new_lines = change.new_text.split("\n")
        if cls._is_full_document_change(lines, change):
            return new_lines
        # keep lines before
        result: list[str] = [] if change.range.start.line <= 0 else lines[0 : change.range.start.line]
        # special case where change covers full lines
        if change.range.start.character <= 0 and change.range.end.character >= len(lines[change.range.end.line]):
            pass
        # special case where change is within 1 line
        elif change.range.start.line == change.range.end.line:
            old_line = lines[change.range.start.line]
            if change.range.start.character > 0:
                new_lines[0] = old_line[0 : change.range.start.character] + new_lines[0]
            if change.range.end.character < len(old_line):
                new_lines[-1] += old_line[change.range.end.character :]
        else:
            if change.range.start.character > 0:
                old_line = lines[change.range.start.line]
                new_lines[0] = old_line[0 : change.range.start.character] + new_lines[0]
            if change.range.end.character < len(lines[change.range.end.line]):
                old_line = lines[change.range.end.line]
                new_lines[-1] += old_line[change.range.end.character :]
        result.extend(new_lines)
        # keep lines after
        if change.range.end.line < len(lines) - 1:
            result.extend(lines[change.range.end.line + 1 :])
        return result

    @classmethod
    def _is_full_document_change(cls, lines: list[str], change: TextEdit) -> bool:
        # A range's end is exclusive. Therefore full document range goes from (0, 0) to (l, 0) where l is the number
        # of lines in the document.
        return (
            change.range.start.line == 0
            and change.range.start.character == 0
            and change.range.end.line >= len(lines)
            and change.range.end.character >= 0
        )


class DiagnosticConverter(abc.ABC):

    _KIND_NAMES = {e.name for e in ErrorKind}

    @classmethod
    def apply(cls, file_path: Path, diagnostic: Diagnostic) -> TranspileError:
        code = str(diagnostic.code)
        kind = ErrorKind.INTERNAL
        parts = code.split("-")
        if len(parts) >= 2 and parts[0] in cls._KIND_NAMES:
            kind = ErrorKind[parts[0]]
            parts.pop(0)
            code = "-".join(parts)
        severity = cls._convert_severity(diagnostic.severity)
        lsp_range = cls._convert_range(diagnostic.range)
        return TranspileError(
            code=code, kind=kind, severity=severity, path=file_path, message=diagnostic.message, range=lsp_range
        )

    @classmethod
    def _convert_range(cls, lsp_range: LSPRange | None) -> CodeRange | None:
        if not lsp_range:
            return None
        return CodeRange(cls._convert_position(lsp_range.start), cls._convert_position(lsp_range.end))

    @classmethod
    def _convert_position(cls, lsp_position: LSPPosition) -> CodePosition:
        return CodePosition(lsp_position.line, lsp_position.character)

    @classmethod
    def _convert_severity(cls, severity: DiagnosticSeverity | None) -> ErrorSeverity:
        if severity == DiagnosticSeverity.Information:
            return ErrorSeverity.INFO
        if severity == DiagnosticSeverity.Warning:
            return ErrorSeverity.WARNING
        if severity == DiagnosticSeverity.Error:
            return ErrorSeverity.ERROR
        return ErrorSeverity.INFO


class LSPEngine(TranspileEngine):

    @classmethod
    def from_config_path(cls, config_path: Path) -> LSPEngine:
        config = LSPConfig.load(config_path)
        return LSPEngine(config_path.parent, config)

    @classmethod
    def client_metadata(cls) -> tuple[str, str]:
        """Obtain the name and version for this LSP client, respectively in a tuple."""
        product_info = ProductInfo.from_class(cls)
        return product_info.product_name(), product_info.version()

    def __init__(self, workdir: Path, config: LSPConfig) -> None:
        self._workdir = workdir
        self._config = config
        name, version = self.client_metadata()
        self._client = _LanguageClient(name, version)
        self._init_response: InitializeResult | None = None

    @property
    def transpiler_name(self) -> str:
        return self._config.name

    def options_for_dialect(self, source_dialect: str) -> Sequence[LSPConfigOptionV1]:
        """Get the options supported when transpiling a given source dialect."""
        return self._config.options_for_dialect(source_dialect)

    @property
    def supported_dialects(self) -> Sequence[str]:
        return self._config.remorph.dialects

    @property
    def server_has_transpile_capability(self) -> bool:
        return self._client.transpile_to_databricks_capability is not None

    async def initialize(self, config: TranspileConfig) -> None:
        if self.is_alive:
            raise IllegalStateException("LSP engine is already initialized")
        try:
            await self._do_initialize(config)
            await self._await_for_transpile_capability()
        # it is good practice to catch broad exceptions raised by launching a child process
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error("LSP initialization failed", exc_info=e)

    async def _do_initialize(self, config: TranspileConfig) -> None:
        await self._start_server()
        input_path = config.input_path
        root_path = input_path if input_path.is_dir() else input_path.parent
        params = InitializeParams(
            capabilities=self._client_capabilities(),
            client_info=ClientInfo(name=self._client.name, version=self._client.version),
            process_id=os.getpid(),
            root_uri=str(root_path.absolute().as_uri()),
            workspace_folders=None,  # for now, we only support a single workspace = root_uri
            initialization_options=self._initialization_options(config),
        )
        logger.debug(f"LSP init params: {params}")
        self._init_response = await self._client.initialize_async(params)

    async def _start_server(self) -> None:
        # Sanity-check and split the command-line into components.
        if not (command_line := self._config.remorph.command_line):
            raise ValueError(f"Missing command line for LSP server: {self._config.path}")
        executable, *args = command_line

        # Extract the environment, preparing to ensure that PATH is set correctly.
        env: dict[str, str] = os.environ | self._config.remorph.env_vars
        path = env.get("PATH", os.defpath)

        # If we have a virtual environment, ensure the bin directory is first on the PATH. This normally takes
        # care of python executables, but also deals with any entry-points that the LSP server might install.
        if (venv_path := self._workdir / ".venv").exists():
            executable, additional_path = self._activate_venv(venv_path, executable)
            # Ensure PATH is in sync with the search path we will use to locate the LSP server executable.
            env["PATH"] = path = f"{additional_path}{os.pathsep}{path}"
        logger.debug(f"Using PATH for launching LSP server: {path}")

        # Locate the LSP server executable in a platform-independent way.
        # Reference: https://docs.python.org/3/library/subprocess.html#popen-constructor
        executable = shutil.which(executable, path=path) or executable

        await self._launch_executable(executable, args, env)

    @staticmethod
    def _activate_venv(venv_path: Path, executable: str) -> tuple[str, Path]:
        """Obtain the bin/script directory for the virtual environment, to extend the search path."""
        logger.debug(f"Detected virtual environment to use at: {venv_path}")
        use_symlinks = sys.platform != "win32"
        builder = venv.EnvBuilder(symlinks=use_symlinks)
        context = builder.ensure_directories(venv_path)

        # Workaround for Windows, where bin_path (Scripts/) doesn't contain python3.exe: if the executable is python
        # or python3, we substitute it for what is needed to launch the venv's python interpreter.
        if os.path.normcase(executable) in {"python", "python3"}:
            executable = context.env_exec_cmd

        return executable, context.bin_path

    async def _launch_executable(self, executable: str, args: Sequence[str], env: Mapping[str, str]) -> None:
        log_level = logging.getLevelName(logging.getLogger("databricks").level)
        # TODO: Remove the --log_level argument once all our transpilers support the environment variable.
        args = [*args, f"--log_level={log_level}"]
        env = {**env, "DATABRICKS_LAKEBRIDGE_LOG_LEVEL": log_level}
        logger.debug(f"Starting LSP engine: {executable} {args} (cwd={self._workdir})")
        await self._client.start_io(executable, *args, env=env, cwd=self._workdir)

    def _client_capabilities(self):
        return ClientCapabilities()  # TODO do we need to refine this ?

    def _initialization_options(self, config: TranspileConfig):
        return {
            "remorph": {
                "source-dialect": config.source_dialect,
            },
            "options": config.transpiler_options,
            "custom": self._config.custom,
        }

    async def _await_for_transpile_capability(self):
        for _ in range(1, 100):
            if self._client.transpile_to_databricks_capability:
                return
            await asyncio.sleep(0.1)
        if not self._client.transpile_to_databricks_capability:
            msg = f"LSP server did not register its {TRANSPILE_TO_DATABRICKS_METHOD} capability"
            raise FeatureRequestError(msg)

    async def shutdown(self):
        await self._client.shutdown_async(None)
        self._client.exit(None)
        await self._client.stop()

    @property
    def is_alive(self):
        return self._client.is_alive

    async def transpile(
        self, source_dialect: str, target_dialect: str, source_code: str, file_path: Path
    ) -> TranspileResult:
        self.open_document(file_path, source_code=source_code)
        response = await self.transpile_document(file_path)
        self.close_document(file_path)
        return ChangeManager.apply(source_code, response.changes, response.diagnostics, file_path)

    def open_document(self, file_path: Path, source_code: str) -> None:
        text_document = TextDocumentItem(
            uri=file_path.absolute().as_uri(), language_id=LanguageKind.Sql, version=1, text=source_code
        )
        params = DidOpenTextDocumentParams(text_document)
        self._client.text_document_did_open(params)

    def close_document(self, file_path: Path) -> None:
        text_document = TextDocumentIdentifier(uri=file_path.absolute().as_uri())
        params = DidCloseTextDocumentParams(text_document)
        self._client.text_document_did_close(params)

    async def transpile_document(self, file_path: Path) -> TranspileDocumentResult:
        params = TranspileDocumentParams(uri=file_path.absolute().as_uri(), language_id=LanguageKind.Sql)
        result = await self._client.transpile_document_async(params)
        return result

    # TODO infer the below from config file
    def is_supported_file(self, file: Path) -> bool:
        if self._is_bladebridge() or self._is_test_transpiler():
            return True
        if self._is_morpheus():
            return is_sql_file(file) or is_dbt_project_file(file)
        # then only support sql
        return is_sql_file(file)

    # TODO remove this
    def _is_test_transpiler(self):
        return self._config.remorph.name == "test-transpiler"

    # TODO remove this
    def _is_bladebridge(self):
        return self._config.remorph.name == "Bladebridge"

    # TODO remove this
    def _is_morpheus(self):
        return self._config.remorph.name == "Morpheus"



from dataclasses import dataclass, field


@dataclass
class Step:
    name: str
    type: str | None
    extract_source: str
    mode: str | None
    frequency: str | None
    flag: str | None
    dependencies: list[str] = field(default_factory=list)
    comment: str | None = None

    def __post_init__(self):
        if self.frequency is None:
            self.frequency = "once"
        if self.flag is None:
            self.flag = "active"
        if self.mode is None:
            self.mode = "append"


@dataclass
class PipelineConfig:
    name: str
    version: str
    extract_folder: str
    comment: str | None = None
    steps: list[Step] = field(default_factory=list)



import logging
from abc import ABC

import sqlglot.expressions as exp
from sqlglot import Dialect, parse_one

from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.exception import InvalidInputException
from databricks.labs.lakebridge.reconcile.query_builder.expression_generator import (
    DataType_transform_mapping,
    transform_expression,
    build_column,
)
from databricks.labs.lakebridge.reconcile.recon_config import Schema, Table, Aggregate
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_dialect, SQLGLOT_DIALECTS

logger = logging.getLogger(__name__)


class QueryBuilder(ABC):
    def __init__(self, table_conf: Table, schema: list[Schema], layer: str, engine: Dialect, data_source: DataSource):
        self._table_conf = table_conf
        self._schema = schema
        self._layer = layer
        self._engine = engine
        self._data_source = data_source

    @property
    def engine(self) -> Dialect:
        return self._engine if self.layer == "source" else get_dialect("databricks")

    @property
    def layer(self) -> str:
        return self._layer

    @property
    def schema(self) -> list[Schema]:
        return self._schema

    @property
    def table_conf(self) -> Table:
        return self._table_conf

    @property
    def select_columns(self) -> set[str]:
        return self.table_conf.get_select_columns(self._schema, self._layer)

    @property
    def threshold_columns(self) -> set[str]:
        return self.table_conf.get_threshold_columns(self._layer)

    @property
    def join_columns(self) -> set[str] | None:
        return self.table_conf.get_join_columns(self._layer)

    @property
    def drop_columns(self) -> set[str]:
        return self._table_conf.get_drop_columns(self._layer)

    @property
    def partition_column(self) -> set[str]:
        return self._table_conf.get_partition_column(self._layer)

    @property
    def filter(self) -> str | None:
        return self._table_conf.get_filter(self._layer)

    @property
    def user_transformations(self) -> dict[str, str]:
        if self._table_conf.transformations:
            if self.layer == "source":
                return {
                    trans.column_name: (
                        trans.source
                        if trans.source
                        else self._data_source.normalize_identifier(trans.column_name).source_normalized
                    )
                    for trans in self._table_conf.transformations
                }
            return {
                self._table_conf.get_layer_src_to_tgt_col_mapping(trans.column_name, self.layer): (
                    trans.target
                    if trans.target
                    else self._table_conf.get_layer_src_to_tgt_col_mapping(trans.column_name, self.layer)
                )
                for trans in self._table_conf.transformations
            }
        return {}

    @property
    def aggregates(self) -> list[Aggregate] | None:
        return self.table_conf.aggregates

    def add_transformations(self, aliases: list[exp.Expression], source: Dialect) -> list[exp.Expression]:
        if self.user_transformations:
            alias_with_user_transforms = self._apply_user_transformation(aliases)
            default_transform_schema: list[Schema] = list(
                filter(lambda sch: sch.column_name not in self.user_transformations.keys(), self.schema)
            )
            return self._apply_default_transformation(alias_with_user_transforms, default_transform_schema, source)
        return self._apply_default_transformation(aliases, self.schema, source)

    def _apply_user_transformation(self, aliases: list[exp.Expression]) -> list[exp.Expression]:
        with_transform = []
        for alias in aliases:
            with_transform.append(alias.transform(self._user_transformer, self.user_transformations))
        return with_transform

    def _user_transformer(self, node: exp.Expression, user_transformations: dict[str, str]) -> exp.Expression:
        if isinstance(node, exp.Column) and user_transformations:
            normalized_column = self._data_source.normalize_identifier(node.name)
            ansi_name = normalized_column.ansi_normalized
            if ansi_name in user_transformations.keys():
                return parse_one(
                    user_transformations.get(ansi_name, normalized_column.source_normalized), read=self.engine
                )
        return node

    def _apply_default_transformation(
        self, aliases: list[exp.Expression], schema: list[Schema], source: Dialect
    ) -> list[exp.Expression]:
        with_transform = []
        for alias in aliases:
            with_transform.append(alias.transform(self._default_transformer, schema, source))
        return with_transform

    def _default_transformer(self, node: exp.Expression, schema: list[Schema], source: Dialect) -> exp.Expression:

        def _get_transform(datatype: str):
            source_dialects = [source_key for source_key, dialect in SQLGLOT_DIALECTS.items() if dialect == source]
            source_dialect = source_dialects[0] if source_dialects else "universal"

            source_mapping = DataType_transform_mapping.get(source_dialect, {})

            if source_mapping.get(datatype.upper()) is not None:
                return source_mapping.get(datatype.upper())
            if source_mapping.get("default") is not None:
                return source_mapping.get("default")

            return DataType_transform_mapping.get("universal", {}).get("default")

        schema_dict = {v.column_name: v.data_type for v in schema}
        if isinstance(node, exp.Column):
            normalized_column = self._data_source.normalize_identifier(node.name)
            ansi_name = normalized_column.ansi_normalized
            if ansi_name in schema_dict.keys():
                transform = _get_transform(schema_dict.get(ansi_name, normalized_column.source_normalized))
                return transform_expression(node, transform)
        return node

    def _validate(self, field: set[str] | list[str] | None, message: str):
        if field is None:
            message = f"Exception for {self.table_conf.target_name} target table in {self.layer} layer --> {message}"
            logger.error(message)
            raise InvalidInputException(message)

    def _build_column_with_alias(self, column: str):
        return build_column(
            this=self._build_column_name_source_normalized(column),
            alias=DialectUtils.unnormalize_identifier(
                self.table_conf.get_layer_tgt_to_src_col_mapping(column, self.layer)
            ),
            quoted=True,
        )

    def _build_column_name_source_normalized(self, column: str):
        return self._data_source.normalize_identifier(column).source_normalized

    def _build_alias_source_normalized(self, column: str):
        return self._data_source.normalize_identifier(
            self.table_conf.get_layer_tgt_to_src_col_mapping(column, self.layer)
        ).source_normalized



from databricks.sdk.core import with_user_agent_extra, with_product
from databricks.labs.blueprint.logger import install_logger
from databricks.labs.lakebridge.__about__ import __version__

install_logger()

# Add lakebridge/<version> for projects depending on lakebridge as a library
with_user_agent_extra("lakebridge", __version__)

# Add lakebridge/<version> for re-packaging of lakebridge, where product name is omitted
with_product("lakebridge", __version__)



import asyncio
import dataclasses
import logging
from email import policy
from email.message import Message
from email.parser import Parser as EmailParser
from pathlib import Path
from typing import cast
import itertools

from databricks.labs.blueprint.installation import JsonObject
from databricks.labs.blueprint.paths import read_text
from databricks.labs.lakebridge.__about__ import __version__
from databricks.labs.lakebridge.config import (
    TranspileConfig,
    TranspileResult,
    ValidationResult,
)
from databricks.labs.lakebridge.helpers import db_sql
from databricks.labs.lakebridge.helpers.execution_time import timeit
from databricks.labs.lakebridge.helpers.file_utils import (
    dir_walk,
    make_dir,
)
from databricks.labs.lakebridge.transpiler.transpile_engine import TranspileEngine
from databricks.labs.lakebridge.transpiler.transpile_status import (
    TranspileStatus,
    TranspileError,
    ErrorKind,
    ErrorSeverity,
)
from databricks.labs.lakebridge.helpers.validation import Validator
from databricks.labs.lakebridge.transpiler.sqlglot.sqlglot_engine import SqlglotEngine
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class TranspilingContext:
    config: TranspileConfig
    validator: Validator | None
    transpiler: TranspileEngine
    input_path: Path
    output_folder: Path
    output_path: Path | None = None
    source_code: str | None = None
    transpiled_code: str | None = None


def _validate_transpiled_sql(context: TranspilingContext, content: str, error_list: list[TranspileError]) -> str:
    if context.validator is None:
        return content
    validation_result = _validation(context.validator, context.config, str(content))
    # Potentially expensive, only evaluate if debug is enabled
    if logger.isEnabledFor(logging.DEBUG):
        msg = f"Finished validating transpiled code for file: {context.input_path} (result: {validation_result})"
        logger.debug(msg)
    if validation_result.exception_msg is not None:
        error = TranspileError(
            "VALIDATION_ERROR",
            ErrorKind.VALIDATION,
            ErrorSeverity.WARNING,
            context.input_path,
            validation_result.exception_msg,
        )
        error_list.append(error)
    return validation_result.validated_sql


async def _process_one_file(context: TranspilingContext) -> tuple[int, list[TranspileError]]:
    input_path = context.input_path

    logger.debug(f"Started processing file: {input_path}")

    if not context.config.source_dialect:
        error = TranspileError(
            code="no-source-dialect-specified",
            kind=ErrorKind.INTERNAL,
            severity=ErrorSeverity.ERROR,
            path=context.input_path,
            message="No source dialect specified",
        )
        return 0, [error]

    # Check if it looks like XML, where we need to sniff the encoding instead of relying on a BOM or defaulting to the
    # local platform encoding.
    sniff_xml_encoding = input_path.suffix.lower() == ".xml"

    source_code = read_text(input_path, detect_xml=sniff_xml_encoding)
    context = dataclasses.replace(context, source_code=source_code)

    transpile_result = await _transpile(
        context.transpiler,
        str(context.config.source_dialect),
        context.config.target_dialect,
        source_code,
        context.input_path,
    )

    # Potentially expensive, only evaluate if debug is enabled
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(f"Finished transpiling file: {context.input_path} (result: {transpile_result})")

    error_list = list(transpile_result.error_list)
    context = dataclasses.replace(context, transpiled_code=transpile_result.transpiled_code)

    output_path = context.output_path
    assert output_path is not None, "Output path must be set in the context"
    output_path.parent.mkdir(exist_ok=True)

    if _is_mime_result(transpile_result):
        _process_mime_result(context, error_list)
    else:
        _process_non_mime_result(context, error_list)

    return transpile_result.success_count, error_list


def _is_mime_result(result: TranspileResult):
    return result.transpiled_code.startswith("Content-Type: multipart/mixed; boundary=")


def _process_mime_result(context: TranspilingContext, error_list: list[TranspileError]) -> None:
    # TODO error handling
    # Added policy to process quoted-printable encoded
    parser = EmailParser(policy=policy.default)
    transpiled_code: str = cast(str, context.transpiled_code)
    message: Message = parser.parsestr(transpiled_code)
    for part in message.walk():
        _process_combined_part(context, part, error_list)


def _process_combined_part(context: TranspilingContext, part: Message, error_list: list[TranspileError]) -> None:
    if part.get_content_type() != "text/plain":
        return  # TODO Need to handle other content types, e.g., text/binary, application/json, etc.
    filename = part.get_filename()
    payload = part.get_payload(decode=True)
    charset = part.get_content_charset() or "utf-8"
    if isinstance(payload, bytes):
        content = payload.decode(charset)
    else:
        content = str(payload)
    logger.debug(f"Processing file: {filename}")

    if not filename:
        return  # TODO Raise exception!!!!
    filename = Path(filename).name
    folder = context.output_folder
    segments = filename.split("/")
    for segment in segments[:-1]:
        folder = folder / segment
        folder.mkdir(parents=True, exist_ok=True)
    output = folder / segments[-1]
    logger.debug(f"Writing output to: {output}")
    # Only validate if output file has .sql suffix
    if output.suffix == ".sql":
        content = _validate_transpiled_sql(context, content, error_list)
    output.write_text(content)


def _process_non_mime_result(context: TranspilingContext, error_list: list[TranspileError]) -> None:

    output_code: str = context.transpiled_code or ""
    output_path = cast(Path, context.output_path)

    if any(err.kind == ErrorKind.PARSING for err in error_list):
        output_code = context.source_code or ""
    elif output_path.suffix == ".sql":
        output_code = _validate_transpiled_sql(context, output_code, error_list)
    with output_path.open("w") as w:
        # The above adds a java-style comment block at the top of the output file
        # This would break .py or .json outputs so we disable it for now.
        # w.write(make_header(context.input_path, error_list))
        w.write(output_code)

    logger.info(f"Processed file: {context.input_path} (errors: {len(error_list)})")


def make_header(file_path: Path, errors: list[TranspileError]) -> str:
    header = ""
    failed_producing_output = False
    diag_by_severity = {
        severity.name: list(diags) for severity, diags in itertools.groupby(errors, key=lambda x: x.severity)
    }
    line_numbers: dict[int, int] = {}

    if ErrorSeverity.ERROR.name in diag_by_severity:
        header += f"/*\n    Failed transpilation of {file_path}\n"
        header += "\n    The following errors were found while transpiling:\n"
        header += _append_diagnostics(diag_by_severity[ErrorSeverity.ERROR.name], line_numbers)
        failed_producing_output = failed_producing_output or any(
            x.kind == ErrorKind.PARSING for x in diag_by_severity[ErrorSeverity.ERROR.name]
        )
    else:
        header += f"/*\n    Successfully transpiled from {file_path}\n"

    if ErrorSeverity.WARNING.name in diag_by_severity:
        header += "\n    The following warnings were found while transpiling:\n"
        header += _append_diagnostics(diag_by_severity[ErrorSeverity.WARNING.name], line_numbers)

    if failed_producing_output:
        header += "\n\n    Parsing errors prevented the converter from translating the input query.\n"
        header += "    We reproduce the input query unchanged below.\n\n"

    header += "*/\n"

    header_line_count = header.count("\n")

    for unshifted in line_numbers:
        line_numbers[unshifted] = header_line_count + unshifted + 1

    return header.format(line_numbers=line_numbers)


def _append_diagnostics(diagnostics: list[TranspileError], line_numbers: dict) -> str:
    header = ""
    grouped_by_message = {msg: list(diags) for msg, diags in itertools.groupby(diagnostics, lambda x: x.message)}
    for msg, occurrences in grouped_by_message.items():
        for occurrence in occurrences:
            if occurrence.range:
                line_numbers.update({occurrence.range.start.line: 0})
        header += _append_diagnostic(msg, occurrences)
    return header


def _append_diagnostic(msg: str, diags: list[TranspileError]) -> str:
    positions = [
        f"[{{line_numbers[{diag.range.start.line}]}}:{diag.range.start.character + 1}]" for diag in diags if diag.range
    ]
    message = msg.replace("{", "{{").replace("}", "}}").replace("/n", "\n        ")
    result = f"      - {message}\n" if len(positions) != 1 else f"      - {positions[0]} {message}\n"
    if len(positions) > 1:
        positions_str = ", ".join(positions)
        result += f"          Occurred {len(diags)} times at the following positions: {positions_str}\n"
    elif len(diags) > 1:
        result += f"          Occurred {len(diags)} times\n"
    return result


async def _process_many_files(
    config: TranspileConfig,
    validator: Validator | None,
    transpiler: TranspileEngine,
    input_path: Path,
    output_folder: Path,
    files: list[Path],
) -> tuple[int, list[TranspileError]]:
    counter = 0
    all_errors: list[TranspileError] = []

    context = TranspilingContext(
        config=config, validator=validator, transpiler=transpiler, input_path=input_path, output_folder=output_folder
    )
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(f"Processing next {len(files)} files: {files}")
    for file in files:
        if file.name.startswith("."):
            logger.debug(f"Ignored invisible file: {file}")
            continue
        if not transpiler.is_supported_file(file):
            logger.debug(f"Ignored file: {file}")
            continue
        context = dataclasses.replace(context, input_path=file, output_path=output_folder / file.name)
        success_count, error_list = await _process_one_file(context)
        counter = counter + success_count
        all_errors.extend(error_list)
    return counter, all_errors


async def _process_input_dir(config: TranspileConfig, validator: Validator | None, transpiler: TranspileEngine):
    error_list = []
    file_list = []
    counter = 0
    input_path = config.input_path
    output_folder = config.output_path
    if output_folder is None:
        output_folder = input_path.parent / "transpiled"
    make_dir(output_folder)
    for source_dir, _, files in dir_walk(input_path):
        relative_path = cast(Path, source_dir).relative_to(input_path)
        transpiled_dir = output_folder / relative_path
        logger.debug(f"Transpiling files from folder: {source_dir} -> {transpiled_dir}")
        file_list.extend(files)
        no_of_sqls, errors = await _process_many_files(config, validator, transpiler, input_path, transpiled_dir, files)
        counter = counter + no_of_sqls
        error_list.extend(errors)
    return TranspileStatus(file_list, counter, error_list)


async def _process_input_file(
    config: TranspileConfig, validator: Validator | None, transpiler: TranspileEngine
) -> TranspileStatus:
    if not transpiler.is_supported_file(config.input_path):
        msg = f"{config.input_source} is not a supported file."
        logger.warning(msg)
        # silently ignore non-sql files
        return TranspileStatus([], 0, [])
    msg = f"Transpiling file: {config.input_path!s}"
    logger.info(msg)
    output_folder = config.output_path
    if output_folder is None:
        output_folder = config.input_path.parent / "transpiled"
    make_dir(output_folder)
    output_file = output_folder / config.input_path.name
    context = TranspilingContext(
        config=config,
        validator=validator,
        transpiler=transpiler,
        input_path=config.input_path,
        output_folder=output_folder,
        output_path=output_file,
    )
    no_of_sqls, error_list = await _process_one_file(context)
    return TranspileStatus([config.input_path], no_of_sqls, error_list)


async def transpile(
    workspace_client: WorkspaceClient, engine: TranspileEngine, config: TranspileConfig
) -> tuple[JsonObject, list[TranspileError]]:
    await engine.initialize(config)
    status, errors = await _do_transpile(workspace_client, engine, config)
    await engine.shutdown()
    logger.info("Done transpiling.")
    return status, errors


async def _do_transpile(
    workspace_client: WorkspaceClient, engine: TranspileEngine, config: TranspileConfig
) -> tuple[JsonObject, list[TranspileError]]:
    """
    [Experimental] Transpiles the SQL queries from one dialect to another.

    :param workspace_client: The WorkspaceClient object.
    :param engine: The TranspileEngine.
    :param config: The configuration for the morph operation.
    """
    if not config.input_source:
        logger.error("Input SQL path is not provided.")
        raise ValueError("Input SQL path is not provided.")

    validator = None
    if not config.skip_validation:
        sql_backend = db_sql.get_sql_backend(workspace_client)
        logger.info(f"SQL Backend used for query validation: {type(sql_backend).__name__}")
        validator = Validator(sql_backend)
    if config.input_source is None:
        raise ValueError("Missing input source!")
    if config.input_path.is_dir():
        logger.debug(f"Starting to process input directory: {config.input_path}")
        result = await _process_input_dir(config, validator, engine)
    elif config.input_path.is_file():
        logger.debug(f"Starting to process input file: {config.input_path}")
        result = await _process_input_file(config, validator, engine)
    else:
        msg = f"{config.input_source} does not exist."
        logger.error(msg)
        raise FileNotFoundError(msg)
    logger.debug(f"Transpiler results: {result}")

    if not config.skip_validation:
        logger.info(f"SQL validation errors: {result.validation_error_count}")

    # TODO: Refactor this so that errors are written while transpiling instead of waiting until the end.
    if result.error_list and config.error_path is not None:
        with config.error_path.open("a", encoding="utf-8") as e:
            e.writelines(f"{err}\n" for err in result.error_list)
        error_log_file = str(config.error_path)
    else:
        error_log_file = None

    status = {
        "total_files_processed": len(result.file_list),
        "total_queries_processed": result.no_of_transpiled_queries,
        "analysis_error_count": result.analysis_error_count,
        "parsing_error_count": result.parsing_error_count,
        "validation_error_count": result.validation_error_count,
        "generation_error_count": result.generation_error_count,
        "error_log_file": error_log_file,
    }
    logger.debug(f"Transpiler Status: {status}")
    return status, result.error_list


def verify_workspace_client(workspace_client: WorkspaceClient) -> WorkspaceClient:
    """
    [Private] Verifies and updates the workspace client configuration.

    TODO: In future refactor this function so it can be used for reconcile module without cross access.
    """

    # Using reflection to set right value for _product_info for telemetry
    product_info = getattr(workspace_client.config, '_product_info', (None, None))
    if product_info[0] != "lakebridge":
        setattr(workspace_client.config, '_product_info', ('lakebridge', __version__))

    return workspace_client


async def _transpile(
    engine: TranspileEngine, from_dialect: str, to_dialect: str, source_code: str, input_path: Path
) -> TranspileResult:
    return await engine.transpile(from_dialect, to_dialect, source_code, input_path)


def _validation(
    validator: Validator,
    config: TranspileConfig,
    sql: str,
) -> ValidationResult:
    return validator.validate_format_result(config, sql)


@timeit
def transpile_sql(
    workspace_client: WorkspaceClient,
    config: TranspileConfig,
    source_sql: str,
) -> tuple[TranspileResult, ValidationResult | None]:
    """[Experimental] Transpile a single SQL query from one dialect to another."""
    ws_client: WorkspaceClient = verify_workspace_client(workspace_client)

    engine: TranspileEngine = SqlglotEngine()

    transpiler_result = asyncio.run(
        _transpile(engine, cast(str, config.source_dialect), config.target_dialect, source_sql, Path("inline_sql"))
    )

    if config.skip_validation:
        return transpiler_result, None

    sql_backend = db_sql.get_sql_backend(ws_client)
    logger.info(f"SQL Backend used for query validation: {type(sql_backend).__name__}")
    validator = Validator(sql_backend)
    return transpiler_result, _validation(validator, config, transpiler_result.transpiled_code)


@timeit
def transpile_column_exp(
    workspace_client: WorkspaceClient,
    config: TranspileConfig,
    expressions: list[str],
) -> list[tuple[TranspileResult, ValidationResult | None]]:
    """[Experimental] Transpile a list of SQL expressions from one dialect to another."""
    config.skip_validation = True
    result = []
    for sql in expressions:
        result.append(transpile_sql(workspace_client, config, sql))
    return result



from sqlglot.dialects.snowflake import Snowflake
from databricks.labs.lakebridge.coverage.commons import sqlglot_run_coverage

if __name__ == "__main__":
    sqlglot_run_coverage(Snowflake, "snowflake")



import logging
from datetime import datetime
from functools import reduce

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col, collect_list, create_map, lit
from pyspark.sql.types import StringType, StructField, StructType
from pyspark.errors import PySparkException
from sqlglot import Dialect

from databricks.labs.lakebridge.config import DatabaseConfig, Table, ReconcileMetadataConfig
from databricks.labs.lakebridge.reconcile.recon_config import TableThresholds
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_key_from_dialect
from databricks.labs.lakebridge.reconcile.exception import (
    WriteToTableException,
    ReadAndWriteWithVolumeException,
    CleanFromVolumeException,
)
from databricks.labs.lakebridge.reconcile.recon_output_config import (
    DataReconcileOutput,
    ReconcileOutput,
    ReconcileProcessDuration,
    ReconcileTableOutput,
    SchemaReconcileOutput,
    StatusOutput,
    ReconcileRecordCount,
    AggregateQueryOutput,
)
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)

_RECON_TABLE_NAME = "main"
_RECON_METRICS_TABLE_NAME = "metrics"
_RECON_DETAILS_TABLE_NAME = "details"
_RECON_AGGREGATE_RULES_TABLE_NAME = "aggregate_rules"
_RECON_AGGREGATE_METRICS_TABLE_NAME = "aggregate_metrics"
_RECON_AGGREGATE_DETAILS_TABLE_NAME = "aggregate_details"


class ReconIntermediatePersist:

    def __init__(self, spark: SparkSession, path: str):
        self.spark = spark
        self.path = path

    def _write_unmatched_df_to_volumes(
        self,
        unmatched_df: DataFrame,
    ) -> None:
        unmatched_df.write.format("parquet").mode("overwrite").save(self.path)

    def _read_unmatched_df_from_volumes(self) -> DataFrame:
        return self.spark.read.format("parquet").load(self.path)

    def clean_unmatched_df_from_volume(self):
        try:
            # TODO: for now we are overwriting the intermediate cache path. We should delete the volume in future
            # workspace_client.dbfs.get_status(path)
            # workspace_client.dbfs.delete(path, recursive=True)
            empty_df = self.spark.createDataFrame([], schema=StructType([StructField("empty", StringType(), True)]))
            empty_df.write.format("parquet").mode("overwrite").save(self.path)
            logger.warning(f"Unmatched DF cleaned up from {self.path} successfully.")
        except PySparkException as e:
            message = f"Error cleaning up unmatched DF from {self.path} volumes --> {e}"
            logger.error(message)
            raise CleanFromVolumeException(message) from e

    def write_and_read_unmatched_df_with_volumes(
        self,
        unmatched_df: DataFrame,
    ) -> DataFrame:
        try:
            self._write_unmatched_df_to_volumes(unmatched_df)
            return self._read_unmatched_df_from_volumes()
        except PySparkException as e:
            message = f"Exception in reading or writing unmatched DF with volumes {self.path} --> {e}"
            logger.error(message)
            raise ReadAndWriteWithVolumeException(message) from e


def _write_df_to_delta(df: DataFrame, table_name: str, mode="append"):
    try:
        df.write.mode(mode).saveAsTable(table_name)
        logger.info(f"Data written to {table_name} successfully.")
    except Exception as e:
        message = f"Error writing data to {table_name}: {e}"
        logger.error(message)
        raise WriteToTableException(message) from e


def generate_final_reconcile_output(
    recon_id: str,
    spark: SparkSession,
    metadata_config: ReconcileMetadataConfig = ReconcileMetadataConfig(),
    local_test_run: bool = False,
) -> ReconcileOutput:
    _db_prefix = "default" if local_test_run else f"{metadata_config.catalog}.{metadata_config.schema}"
    recon_df = spark.sql(
        f"""
    SELECT
    CASE
        WHEN COALESCE(MAIN.SOURCE_TABLE.CATALOG, '') <> '' THEN CONCAT(MAIN.SOURCE_TABLE.CATALOG, '.', MAIN.SOURCE_TABLE.SCHEMA, '.', MAIN.SOURCE_TABLE.TABLE_NAME)
        ELSE CONCAT(MAIN.SOURCE_TABLE.SCHEMA, '.', MAIN.SOURCE_TABLE.TABLE_NAME)
    END AS SOURCE_TABLE,
    CONCAT(MAIN.TARGET_TABLE.CATALOG, '.', MAIN.TARGET_TABLE.SCHEMA, '.', MAIN.TARGET_TABLE.TABLE_NAME) AS TARGET_TABLE,
    CASE WHEN lower(MAIN.report_type) in ('all', 'row', 'data') THEN
    CASE
        WHEN METRICS.recon_metrics.row_comparison.missing_in_source = 0 AND METRICS.recon_metrics.row_comparison.missing_in_target = 0 THEN TRUE
        ELSE FALSE
    END
    ELSE NULL END AS ROW,
    CASE WHEN lower(MAIN.report_type) in ('all', 'data') THEN
    CASE
        WHEN (METRICS.run_metrics.status = true) or
         (METRICS.recon_metrics.column_comparison.absolute_mismatch = 0 AND METRICS.recon_metrics.column_comparison.threshold_mismatch = 0 AND METRICS.recon_metrics.column_comparison.mismatch_columns = '') THEN TRUE
        ELSE FALSE
    END
    ELSE NULL END AS COLUMN,
    CASE WHEN lower(MAIN.report_type) in ('all', 'schema') THEN
    CASE
        WHEN METRICS.recon_metrics.schema_comparison = true THEN TRUE
        ELSE FALSE
    END
    ELSE NULL END AS SCHEMA,
    METRICS.run_metrics.exception_message AS EXCEPTION_MESSAGE
    FROM
        {_db_prefix}.{_RECON_TABLE_NAME} MAIN
    INNER JOIN
        {_db_prefix}.{_RECON_METRICS_TABLE_NAME} METRICS
    ON
        (MAIN.recon_table_id = METRICS.recon_table_id)
    WHERE
        MAIN.recon_id = '{recon_id}'
    """
    )
    table_output = []
    for row in recon_df.collect():
        if row.EXCEPTION_MESSAGE is not None and row.EXCEPTION_MESSAGE != "":
            table_output.append(
                ReconcileTableOutput(
                    target_table_name=row.TARGET_TABLE,
                    source_table_name=row.SOURCE_TABLE,
                    status=StatusOutput(),
                    exception_message=row.EXCEPTION_MESSAGE,
                )
            )
        else:
            table_output.append(
                ReconcileTableOutput(
                    target_table_name=row.TARGET_TABLE,
                    source_table_name=row.SOURCE_TABLE,
                    status=StatusOutput(row=row.ROW, column=row.COLUMN, schema=row.SCHEMA),
                    exception_message=row.EXCEPTION_MESSAGE,
                )
            )
    final_reconcile_output = ReconcileOutput(recon_id=recon_id, results=table_output)
    logger.info(f"Final reconcile output: {final_reconcile_output}")
    return final_reconcile_output


def generate_final_reconcile_aggregate_output(
    recon_id: str,
    spark: SparkSession,
    metadata_config: ReconcileMetadataConfig = ReconcileMetadataConfig(),
    local_test_run: bool = False,
) -> ReconcileOutput:
    _db_prefix = "default" if local_test_run else f"{metadata_config.catalog}.{metadata_config.schema}"
    recon_df = spark.sql(
        f"""
        SELECT source_table,
         target_table,
          EVERY(status) AS status,
           ARRAY_JOIN(COLLECT_SET(exception_message), '\n') AS exception_message
        FROM
        (SELECT
            IF(ISNULL(main.source_table.catalog)
                , CONCAT_WS('.', main.source_table.schema, main.source_table.table_name)
                , CONCAT_WS('.', main.source_table.catalog, main.source_table.schema, main.source_table.table_name)) AS source_table,
            CONCAT_WS('.', main.target_table.catalog, main.target_table.schema, main.target_table.table_name) AS target_table,
            IF(metrics.run_metrics.status='true', TRUE , FALSE) AS status,
            metrics.run_metrics.exception_message AS exception_message
            FROM
                {_db_prefix}.{_RECON_TABLE_NAME} main
            INNER JOIN
                {_db_prefix}.{_RECON_AGGREGATE_METRICS_TABLE_NAME} metrics
            ON
                (MAIN.recon_table_id = METRICS.recon_table_id
                AND MAIN.operation_name = 'aggregates-reconcile')
            WHERE
                MAIN.recon_id = '{recon_id}'
        )
        GROUP BY source_table, target_table;
    """
    )
    table_output = []
    for row in recon_df.collect():
        if row.exception_message is not None and row.exception_message != "":
            table_output.append(
                ReconcileTableOutput(
                    target_table_name=row.target_table,
                    source_table_name=row.source_table,
                    status=StatusOutput(),
                    exception_message=row.exception_message,
                )
            )
        else:
            table_output.append(
                ReconcileTableOutput(
                    target_table_name=row.target_table,
                    source_table_name=row.source_table,
                    status=StatusOutput(aggregate=row.status),
                    exception_message=row.exception_message,
                )
            )
    final_reconcile_output = ReconcileOutput(recon_id=recon_id, results=table_output)
    logger.info(f"Final reconcile output: {final_reconcile_output}")
    return final_reconcile_output


class ReconCapture:

    def __init__(
        self,
        database_config: DatabaseConfig,
        recon_id: str,
        report_type: str,
        source_dialect: Dialect,
        ws: WorkspaceClient,
        spark: SparkSession,
        metadata_config: ReconcileMetadataConfig = ReconcileMetadataConfig(),
        local_test_run: bool = False,
    ):
        self.database_config = database_config
        self.recon_id = recon_id
        self.report_type = report_type
        self.source_dialect = source_dialect
        self.ws = ws
        self.spark = spark
        self._db_prefix = "default" if local_test_run else f"{metadata_config.catalog}.{metadata_config.schema}"

    def _generate_recon_main_id(
        self,
        table_conf: Table,
    ) -> int:
        full_source_table = (
            f"{self.database_config.source_schema}.{table_conf.source_name}"
            if self.database_config.source_catalog is None
            else f"{self.database_config.source_catalog}.{self.database_config.source_schema}.{table_conf.source_name}"
        )
        full_target_table = (
            f"{self.database_config.target_catalog}.{self.database_config.target_schema}.{table_conf.target_name}"
        )
        return hash(f"{self.recon_id}{full_source_table}{full_target_table}")

    def _insert_into_main_table(
        self,
        recon_table_id: int,
        table_conf: Table,
        recon_process_duration: ReconcileProcessDuration,
        operation_name: str = "reconcile",
    ) -> None:
        source_dialect_key = get_key_from_dialect(self.source_dialect)
        df = self.spark.sql(
            f"""
                select {recon_table_id} as recon_table_id,
                '{self.recon_id}' as recon_id,
                case
                    when '{source_dialect_key}' = 'databricks' then 'Databricks'
                    when '{source_dialect_key}' = 'snowflake' then 'Snowflake'
                    when '{source_dialect_key}' = 'oracle' then 'Oracle'
                    else '{source_dialect_key}'
                end as source_type,
                named_struct(
                    'catalog', case when '{self.database_config.source_catalog}' = 'None' then null else '{self.database_config.source_catalog}' end,
                    'schema', '{self.database_config.source_schema}',
                    'table_name', '{table_conf.source_name}'
                ) as source_table,
                named_struct(
                    'catalog', '{self.database_config.target_catalog}',
                    'schema', '{self.database_config.target_schema}',
                    'table_name', '{table_conf.target_name}'
                ) as target_table,
                '{self.report_type}' as report_type,
                '{operation_name}' as operation_name,
                cast('{recon_process_duration.start_ts}' as timestamp) as start_ts,
                cast('{recon_process_duration.end_ts}' as timestamp) as end_ts
            """
        )
        _write_df_to_delta(df, f"{self._db_prefix}.{_RECON_TABLE_NAME}")

    @classmethod
    def _is_mismatch_within_threshold_limits(
        cls, data_reconcile_output: DataReconcileOutput, table_conf: Table, record_count: ReconcileRecordCount
    ):
        total_mismatch_count = (
            data_reconcile_output.mismatch_count + data_reconcile_output.threshold_output.threshold_mismatch_count
        )
        logger.info(f"total_mismatch_count : {total_mismatch_count}")
        logger.warning(f"reconciled_record_count : {record_count}")
        # if the mismatch count is 0 then no need of checking bounds.
        if total_mismatch_count == 0:
            return True
        # pull out table thresholds
        thresholds: list[TableThresholds] = (
            [threshold for threshold in table_conf.table_thresholds if threshold.model == "mismatch"]
            if table_conf.table_thresholds
            else []
        )
        # if not table thresholds are provided return false
        if not thresholds:
            return False

        res = None
        for threshold in thresholds:
            mode = threshold.get_mode()
            lower_bound = int(threshold.lower_bound.replace("%", ""))
            upper_bound = int(threshold.upper_bound.replace("%", ""))
            if mode == "absolute":
                res = lower_bound <= total_mismatch_count <= upper_bound
            if mode == "percentage":
                lower_bound = int(round((lower_bound / 100) * record_count.source))
                upper_bound = int(round((upper_bound / 100) * record_count.source))
                res = lower_bound <= total_mismatch_count <= upper_bound

        return res

    def _insert_into_metrics_table(
        self,
        recon_table_id: int,
        data_reconcile_output: DataReconcileOutput,
        schema_reconcile_output: SchemaReconcileOutput,
        table_conf: Table,
        record_count: ReconcileRecordCount,
    ) -> None:
        status = False
        if data_reconcile_output.exception in {None, ''} and schema_reconcile_output.exception in {None, ''}:
            status = (
                # validate for both exact mismatch and threshold mismatch
                self._is_mismatch_within_threshold_limits(
                    data_reconcile_output=data_reconcile_output, table_conf=table_conf, record_count=record_count
                )
                and data_reconcile_output.missing_in_src_count == 0
                and data_reconcile_output.missing_in_tgt_count == 0
                and schema_reconcile_output.is_valid
            )

        exception_msg = ""
        if schema_reconcile_output.exception is not None:
            exception_msg = schema_reconcile_output.exception.replace("'", '').replace('"', '')
        if data_reconcile_output.exception is not None:
            exception_msg = data_reconcile_output.exception.replace("'", '').replace('"', '')

        insertion_time = str(datetime.now())
        mismatch_columns = []
        if data_reconcile_output.mismatch and data_reconcile_output.mismatch.mismatch_columns:
            mismatch_columns = data_reconcile_output.mismatch.mismatch_columns

        df = self.spark.sql(
            f"""
                select {recon_table_id} as recon_table_id,
                named_struct(
                    'row_comparison', case when '{self.report_type.lower()}' in ('all', 'row', 'data')
                        and '{exception_msg}' = '' then
                     named_struct(
                        'missing_in_source', cast({data_reconcile_output.missing_in_src_count} as bigint),
                        'missing_in_target', cast({data_reconcile_output.missing_in_tgt_count} as bigint)
                    ) else null end,
                    'column_comparison', case when '{self.report_type.lower()}' in ('all', 'data')
                        and '{exception_msg}' = '' then
                    named_struct(
                        'absolute_mismatch', cast({data_reconcile_output.mismatch_count} as bigint),
                        'threshold_mismatch', cast({data_reconcile_output.threshold_output.threshold_mismatch_count} as bigint),
                        'mismatch_columns', '{",".join(mismatch_columns)}'
                    ) else null end,
                    'schema_comparison', case when '{self.report_type.lower()}' in ('all', 'schema')
                        and '{exception_msg}' = '' then
                        {schema_reconcile_output.is_valid} else null end
                ) as recon_metrics,
                named_struct(
                    'status', {status},
                    'run_by_user', '{self.ws.current_user.me().user_name}',
                    'exception_message', "{exception_msg}"
                ) as run_metrics,
                cast('{insertion_time}' as timestamp) as inserted_ts
            """
        )
        _write_df_to_delta(df, f"{self._db_prefix}.{_RECON_METRICS_TABLE_NAME}")

    @classmethod
    def _create_map_column(
        cls,
        recon_table_id: int,
        df: DataFrame,
        recon_type: str,
        status: bool,
    ) -> DataFrame:
        columns = df.columns
        # Create a list of column names and their corresponding column values
        map_args = []
        for column in columns:
            map_args.extend([lit(column).alias(column + "_key"), col(column).cast("string").alias(column + "_value")])
        # Create a new DataFrame with a map column
        df = df.select(create_map(*map_args).alias("data"))
        df = (
            df.withColumn("recon_table_id", lit(recon_table_id))
            .withColumn("recon_type", lit(recon_type))
            .withColumn("status", lit(status))
            .withColumn("inserted_ts", lit(datetime.now()))
        )
        return (
            df.groupBy("recon_table_id", "recon_type", "status", "inserted_ts")
            .agg(collect_list("data").alias("data"))
            .selectExpr("recon_table_id", "recon_type", "status", "data", "inserted_ts")
        )

    def _create_map_column_and_insert(
        self,
        recon_table_id: int,
        df: DataFrame,
        recon_type: str,
        status: bool,
    ) -> None:
        df = self._create_map_column(recon_table_id, df, recon_type, status)
        _write_df_to_delta(df, f"{self._db_prefix}.{_RECON_DETAILS_TABLE_NAME}")

    def _insert_into_details_table(
        self,
        recon_table_id: int,
        reconcile_output: DataReconcileOutput,
        schema_output: SchemaReconcileOutput,
    ):
        if reconcile_output.mismatch_count > 0 and reconcile_output.mismatch.mismatch_df:
            self._create_map_column_and_insert(
                recon_table_id,
                reconcile_output.mismatch.mismatch_df,
                "mismatch",
                False,
            )

        if reconcile_output.missing_in_src_count > 0 and reconcile_output.missing_in_src:
            self._create_map_column_and_insert(
                recon_table_id,
                reconcile_output.missing_in_src,
                "missing_in_source",
                False,
            )

        if reconcile_output.missing_in_tgt_count > 0 and reconcile_output.missing_in_tgt:
            self._create_map_column_and_insert(
                recon_table_id,
                reconcile_output.missing_in_tgt,
                "missing_in_target",
                False,
            )

        if (
            reconcile_output.threshold_output.threshold_mismatch_count > 0
            and reconcile_output.threshold_output.threshold_df
        ):
            self._create_map_column_and_insert(
                recon_table_id,
                reconcile_output.threshold_output.threshold_df,
                "threshold_mismatch",
                False,
            )

        if schema_output.compare_df is not None:
            self._create_map_column_and_insert(
                recon_table_id, schema_output.compare_df, "schema", schema_output.is_valid
            )

    def _get_df(
        self,
        recon_table_id: int,
        agg_data: DataReconcileOutput,
        recon_type: str,
    ):

        column_count = agg_data.mismatch_count
        agg_df = agg_data.mismatch.mismatch_df
        match recon_type:
            case "missing_in_source":
                column_count = agg_data.missing_in_src_count
                agg_df = agg_data.missing_in_src
            case "missing_in_target":
                column_count = agg_data.missing_in_tgt_count
                agg_df = agg_data.missing_in_tgt

        if column_count > 0 and agg_df:
            return self._create_map_column(
                recon_table_id,
                agg_df,
                recon_type,
                False,
            )
        return None

    @classmethod
    def _union_dataframes(cls, df_list: list[DataFrame]) -> DataFrame:
        return reduce(lambda agg_df, df: agg_df.unionByName(df), df_list)

    def _insert_aggregates_into_metrics_table(
        self,
        recon_table_id: int,
        reconcile_agg_output_list: list[AggregateQueryOutput],
    ) -> None:

        agg_metrics_df_list = []
        for agg_output in reconcile_agg_output_list:
            agg_data = agg_output.reconcile_output

            status = False
            if agg_data.exception in {None, ''}:
                status = not (
                    agg_data.mismatch_count > 0
                    or agg_data.missing_in_src_count > 0
                    or agg_data.missing_in_tgt_count > 0
                )

            exception_msg = ""
            if agg_data.exception is not None:
                exception_msg = agg_data.exception.replace("'", '').replace('"', '')

            insertion_time = str(datetime.now())

            # If there is any exception while running the Query,
            # each rule is stored, with the Exception message in the metrics table
            assert agg_output.rule, "Aggregate Rule must be present for storing the metrics"
            rule_id = hash(f"{recon_table_id}_{agg_output.rule.column_from_rule}")

            agg_metrics_df = self.spark.sql(
                f"""
                    select {recon_table_id} as recon_table_id,
                    {rule_id}  as rule_id,
                    if('{exception_msg}' = '', named_struct(
                            'missing_in_source', {agg_data.missing_in_src_count},
                            'missing_in_target', {agg_data.missing_in_tgt_count},
                            'mismatch', {agg_data.mismatch_count}
                    ), null) as recon_metrics,
                    named_struct(
                        'status', {status},
                        'run_by_user', '{self.ws.current_user.me().user_name}',
                        'exception_message', "{exception_msg}"
                    ) as run_metrics,
                    cast('{insertion_time}' as timestamp) as inserted_ts
                """
            )
            agg_metrics_df_list.append(agg_metrics_df)

        agg_metrics_table_df = self._union_dataframes(agg_metrics_df_list)
        _write_df_to_delta(agg_metrics_table_df, f"{self._db_prefix}.{_RECON_AGGREGATE_METRICS_TABLE_NAME}")

    def _insert_aggregates_into_details_table(
        self, recon_table_id: int, reconcile_agg_output_list: list[AggregateQueryOutput]
    ):
        agg_details_df_list = []
        for agg_output in reconcile_agg_output_list:
            agg_details_rule_df_list = []

            mismatch_df = self._get_df(recon_table_id, agg_output.reconcile_output, "mismatch")
            if mismatch_df and not mismatch_df.isEmpty():
                agg_details_rule_df_list.append(mismatch_df)

            missing_src_df = self._get_df(recon_table_id, agg_output.reconcile_output, "missing_in_source")
            if missing_src_df and not missing_src_df.isEmpty():
                agg_details_rule_df_list.append(missing_src_df)

            missing_tgt_df = self._get_df(recon_table_id, agg_output.reconcile_output, "missing_in_target")
            if missing_tgt_df and not missing_tgt_df.isEmpty():
                agg_details_rule_df_list.append(missing_tgt_df)

            if agg_details_rule_df_list:
                agg_details_rule_df = self._union_dataframes(agg_details_rule_df_list)
                if agg_output.rule:
                    rule_id = hash(f"{recon_table_id}_{agg_output.rule.column_from_rule}")
                    agg_details_rule_df = agg_details_rule_df.withColumn("rule_id", lit(rule_id)).select(
                        "recon_table_id", "rule_id", "recon_type", "data", "inserted_ts"
                    )
                    agg_details_df_list.append(agg_details_rule_df)
            else:
                logger.warning("Aggregate Details Rules are empty")

        if agg_details_df_list:
            agg_details_table_df = self._union_dataframes(agg_details_df_list)
            _write_df_to_delta(agg_details_table_df, f"{self._db_prefix}.{_RECON_AGGREGATE_DETAILS_TABLE_NAME}")

    def start(
        self,
        data_reconcile_output: DataReconcileOutput,
        schema_reconcile_output: SchemaReconcileOutput,
        table_conf: Table,
        recon_process_duration: ReconcileProcessDuration,
        record_count: ReconcileRecordCount,
    ) -> None:
        recon_table_id = self._generate_recon_main_id(table_conf)
        self._insert_into_main_table(recon_table_id, table_conf, recon_process_duration)
        self._insert_into_metrics_table(
            recon_table_id, data_reconcile_output, schema_reconcile_output, table_conf, record_count
        )
        self._insert_into_details_table(recon_table_id, data_reconcile_output, schema_reconcile_output)

    def store_aggregates_metrics(
        self,
        table_conf: Table,
        recon_process_duration: ReconcileProcessDuration,
        reconcile_agg_output_list: list[AggregateQueryOutput],
    ) -> None:
        recon_table_id = self._generate_recon_main_id(table_conf)
        self._insert_into_main_table(recon_table_id, table_conf, recon_process_duration, 'aggregates-reconcile')
        self._insert_into_rules_table(recon_table_id, reconcile_agg_output_list)
        self._insert_aggregates_into_metrics_table(recon_table_id, reconcile_agg_output_list)
        self._insert_aggregates_into_details_table(
            recon_table_id,
            reconcile_agg_output_list,
        )

    def _insert_into_rules_table(self, recon_table_id: int, reconcile_agg_output_list: list[AggregateQueryOutput]):

        rule_df_list = []
        for agg_output in reconcile_agg_output_list:
            if not agg_output.rule:
                logger.error("Aggregate Rule must be present for storing the rules")
                continue
            rule_id = hash(f"{recon_table_id}_{agg_output.rule.column_from_rule}")
            rule_query = agg_output.rule.get_rule_query(rule_id)
            rule_df_list.append(
                self.spark.sql(rule_query)
                .withColumn("inserted_ts", lit(datetime.now()))
                .select("rule_id", "rule_type", "rule_info", "inserted_ts")
            )

        if rule_df_list:
            rules_table_df = self._union_dataframes(rule_df_list)
            _write_df_to_delta(rules_table_df, f"{self._db_prefix}.{_RECON_AGGREGATE_RULES_TABLE_NAME}")



from datetime import datetime

from pyspark.sql import SparkSession
from databricks.sdk import WorkspaceClient

from databricks.labs.lakebridge.config import ReconcileConfig, TableRecon
from databricks.labs.lakebridge.reconcile import utils
from databricks.labs.lakebridge.reconcile.exception import DataSourceRuntimeException, ReconciliationException
from databricks.labs.lakebridge.reconcile.recon_capture import (
    ReconIntermediatePersist,
    generate_final_reconcile_aggregate_output,
)
from databricks.labs.lakebridge.reconcile.recon_config import AGG_RECONCILE_OPERATION_NAME
from databricks.labs.lakebridge.reconcile.recon_output_config import (
    ReconcileProcessDuration,
    AggregateQueryOutput,
    DataReconcileOutput,
)
from databricks.labs.lakebridge.reconcile.trigger_recon_service import TriggerReconService


class TriggerReconAggregateService:
    @staticmethod
    def trigger_recon_aggregates(
        ws: WorkspaceClient,
        spark: SparkSession,
        table_recon: TableRecon,
        reconcile_config: ReconcileConfig,
        local_test_run: bool = False,
    ):
        reconciler, recon_capture = TriggerReconService.create_recon_dependencies(
            ws, spark, reconcile_config, local_test_run
        )

        # Get the Aggregated Reconciliation Output for each table
        for table_conf in table_recon.tables:
            recon_process_duration = ReconcileProcessDuration(start_ts=str(datetime.now()), end_ts=None)
            try:
                src_schema, tgt_schema = TriggerReconService.get_schemas(
                    reconciler.source, reconciler.target, table_conf, reconcile_config.database_config, False
                )
            except DataSourceRuntimeException as e:
                raise ReconciliationException(message=str(e)) from e

            assert table_conf.aggregates, "Aggregates must be defined for Aggregates Reconciliation"

            try:
                table_reconcile_agg_output_list = reconciler.reconcile_aggregates(table_conf, src_schema, tgt_schema)
            except DataSourceRuntimeException as e:
                table_reconcile_agg_output_list = [
                    AggregateQueryOutput(reconcile_output=DataReconcileOutput(exception=str(e)), rule=None)
                ]

            recon_process_duration.end_ts = str(datetime.now())

            # Persist the data to the delta tables
            recon_capture.store_aggregates_metrics(
                reconcile_agg_output_list=table_reconcile_agg_output_list,
                table_conf=table_conf,
                recon_process_duration=recon_process_duration,
            )

            (
                ReconIntermediatePersist(
                    spark=spark,
                    path=utils.generate_volume_path(table_conf, reconcile_config.metadata_config),
                ).clean_unmatched_df_from_volume()
            )

        return TriggerReconService.verify_successful_reconciliation(
            generate_final_reconcile_aggregate_output(
                recon_id=recon_capture.recon_id,
                spark=spark,
                metadata_config=reconcile_config.metadata_config,
                local_test_run=local_test_run,
            ),
            operation_name=AGG_RECONCILE_OPERATION_NAME,
        )



from dataclasses import dataclass

from sqlglot import expressions as exp
from sqlglot.expressions import AggFunc, Condition, Expression, Func


class NthValue(AggFunc):
    arg_types = {"this": True, "offset": False}


class Parameter(Expression):
    arg_types = {"this": True, "wrapped": False, "suffix": False}


class Collate(Func):
    arg_types = {"this": True, "expressions": True}


class Bracket(Condition):
    arg_types = {"this": True, "expressions": True}


class Split(Func):
    """
    Redefined Split(sqlglot/expression) class with expression: False to handle default delimiter
    Please refer the test case `test_strtok_to_array` -> `select STRTOK_TO_ARRAY('my text is divided')`
    """

    arg_types = {"this": True, "expression": False, "limit": False}


class MakeDate(Func):
    arg_types = {"this": True, "expression": False, "zone": False}


class ConvertTimeZone(Func):
    arg_types = {"srcTZ": True, "tgtTZ": True, "this": False}


class TryToDate(Func):
    arg_types = {"this": True, "format": False}


class TryToTimestamp(Func):
    arg_types = {"this": True, "format": False}


class SplitPart(Func):
    arg_types = {"this": True, "expression": False, "partNum": False}


class StrTok(Func):
    arg_types = {"this": True, "expression": False, "partNum": False}


class TryToNumber(Func):
    arg_types = {"this": True, "expression": False, "precision": False, "scale": False}

    _sql_names = ["TRY_TO_DECIMAL", "TRY_TO_NUMBER", "TRY_TO_NUMERIC"]


class DateFormat(Func):
    arg_types = {"this": True, "expression": False}


class IsInteger(Func):
    pass


class JsonExtractPathText(Func):
    arg_types = {"this": True, "path_name": True}


class BitOr(AggFunc):
    pass


class ArrayConstructCompact(Func):
    arg_types = {"expressions": False}

    is_var_len_args = True


class ArrayIntersection(Func):
    arg_types = {"this": True, "expression": True}


class ArraySlice(Func):
    arg_types = {"this": True, "from": True, "to": True}


class ObjectKeys(Func):
    arg_types = {"this": True}


class ToBoolean(Func):
    arg_types = {"this": True, "raise_error": False}


class ToDouble(Func):
    pass


class ToObject(Func):
    pass


class ToNumber(Func):
    arg_types = {"this": True, "expression": False, "precision": False, "scale": False}

    _sql_names = ["TO_DECIMAL", "TO_NUMBER", "TO_NUMERIC"]


class TimestampFromParts(Func):
    arg_types = {
        "this": True,
        "expression": True,
        "day": True,
        "hour": True,
        "min": True,
        "sec": True,
        "nanosec": False,
        "Zone": False,
    }


class ToVariant(Func):
    pass


class UUID(Func):
    arg_types = {"this": False, "name": False}


class DateTrunc(Func):
    arg_types = {"unit": False, "this": True, "zone": False}


class Median(Func):
    arg_types = {"this": True}


class CumeDist(Func):
    arg_types = {"this": False}


class DenseRank(Func):
    arg_types = {"this": False}


class Rank(Func):
    arg_types = {"this": False}


class PercentRank(Func):
    arg_types = {"this": False}


class Ntile(Func):
    arg_types = {"this": True, "is_string": False}


class ToArray(Func):
    arg_types = {"this": True, "expression": False}


@dataclass
class WithinGroupParams:
    agg_col: exp.Column
    order_cols: list[tuple[exp.Column, bool]]  # List of (column, is ascending)


@dataclass
class AliasInfo:
    name: str
    expression: exp.Expression
    is_same_name_as_column: bool


class MapKeys(Func):
    arg_types = {"this": True}


class ArrayExists(Func):
    arg_types = {"this": True, "expression": True}


class Locate(Func):
    arg_types = {"substring": True, "this": True, "position": False}


class NamedStruct(Func):
    arg_types = {"expressions": True}


class GetJsonObject(Func):
    arg_types = {"this": True, "path": True}



import logging
from pathlib import Path

from databricks.labs.blueprint.paths import read_text

from databricks.labs.lakebridge.helpers.file_utils import get_sql_file, is_sql_file
from databricks.labs.lakebridge.intermediate.dag import DAG

from databricks.labs.lakebridge.transpiler.sqlglot.sqlglot_engine import SqlglotEngine

logger = logging.getLogger(__name__)


class RootTableAnalyzer:

    def __init__(self, engine: SqlglotEngine, source_dialect: str, input_path: Path):
        self.engine = engine
        self.source_dialect = source_dialect
        self.input_path = input_path

    def generate_lineage_dag(self) -> DAG:
        dag = DAG()

        # when input is sql file then parse the file
        if is_sql_file(self.input_path):
            logger.debug(f"Generating Lineage file: {self.input_path}")
            sql_content = read_text(self.input_path)
            self._populate_dag(sql_content, self.input_path, dag)
            return dag  # return after processing the file

        # when the input is a directory
        for path in get_sql_file(self.input_path):
            logger.debug(f"Generating Lineage file: {path}")
            sql_content = read_text(path)
            self._populate_dag(sql_content, path, dag)

        return dag

    def _populate_dag(self, sql_content: str, path: Path, dag: DAG):
        for root_table, child in self.engine.analyse_table_lineage(self.source_dialect, sql_content, path):
            dag.add_node(child)
            dag.add_edge(root_table, child)



# pylint: disable=invalid-name
import logging

from databricks.labs.blueprint.installation import Installation
from databricks.sdk import WorkspaceClient

from databricks.labs.lakebridge.contexts.application import ApplicationContext
from databricks.labs.lakebridge.deployment.upgrade_common import (
    current_table_columns,
    installed_table_columns,
    check_table_mismatch,
    recreate_table_sql,
)
from databricks.labs.lakebridge.helpers import db_sql

logger = logging.getLogger(__name__)


def _upgrade_reconcile_metadata_metrics_table(
    installation: Installation, ws: WorkspaceClient, app_context: ApplicationContext
):
    reconcile_config = app_context.recon_config
    assert reconcile_config, "Reconcile config must be present to upgrade the reconcile metadata main table"
    table_name = "metrics"
    table_identifier = (
        f"{reconcile_config.metadata_config.catalog}.{reconcile_config.metadata_config.schema}.{table_name}"
    )
    installed_columns = installed_table_columns(ws, table_identifier)
    current_columns = current_table_columns(table_name, table_identifier)
    sqls: list | None = [
        f"ALTER TABLE {table_identifier} SET TBLPROPERTIES ('delta.enableTypeWidening' = 'true')",
        f"ALTER TABLE {table_identifier} ALTER COLUMN recon_metrics.row_comparison.missing_in_source TYPE BIGINT",
        f"ALTER TABLE {table_identifier} ALTER COLUMN recon_metrics.row_comparison.missing_in_target TYPE BIGINT",
        f"ALTER TABLE {table_identifier} ALTER COLUMN recon_metrics.column_comparison.absolute_mismatch TYPE BIGINT",
        f"ALTER TABLE {table_identifier} ALTER COLUMN recon_metrics.column_comparison.threshold_mismatch TYPE BIGINT",
    ]
    if check_table_mismatch(installed_columns, current_columns):
        logger.info("Recreating main table")
        sqls = [recreate_table_sql(table_identifier, installed_columns, current_columns, app_context.prompts)]
    if sqls:
        for sql in sqls:
            logger.debug(f"Executing SQL to upgrade metrics table: \n{sql}")
            db_sql.get_sql_backend(ws).execute(sql)
        installation.save(reconcile_config)
        logger.debug("Upgraded Reconcile metrics table")


def upgrade(installation: Installation, ws: WorkspaceClient):
    app_context = ApplicationContext(ws)
    if app_context.recon_config is not None:
        _upgrade_reconcile_metadata_metrics_table(installation, ws, app_context)
